import math

import numpy as np

import orca_gym.adapters.robosuite.utils.transform_utils as T
from orca_gym.adapters.robosuite.controllers.base_controller import Controller
from orca_gym.adapters.robosuite.utils.control_utils import *

# Supported impedance modes
IMPEDANCE_MODES = {"fixed", "variable", "variable_kp"}

# TODO: Maybe better naming scheme to differentiate between input / output min / max and pos/ori limits, etc.


class OperationalSpaceController(Controller):
    """
    Controller for controlling robot arm via operational space control. Allows position and / or orientation control
    of the robot's end effector. For detailed information as to the mathematical foundation for this controller, please
    reference http://khatib.stanford.edu/publications/pdfs/Khatib_1987_RA.pdf

    NOTE: Control input actions can either be taken to be relative to the current position / orientation of the
    end effector or absolute values. In either case, a given action to this controller is assumed to be of the form:
    (x, y, z, ax, ay, az) if controlling pos and ori or simply (x, y, z) if only controlling pos

    Args:
        sim (MjSim): Simulator instance this controller will pull robot state updates from

        eef_name (str): Name of controlled robot arm's end effector (from robot XML)

        joint_indexes (dict): Each key contains sim reference indexes to relevant robot joint information, namely:

            :`'joints'`: list of indexes to relevant robot joints
            :`'qpos'`: list of indexes to relevant robot joint positions
            :`'qvel'`: list of indexes to relevant robot joint velocities

        actuator_range (2-tuple of array of float): 2-Tuple (low, high) representing the robot joint actuator range

        input_max (float or Iterable of float): Maximum above which an inputted action will be clipped. Can be either be
            a scalar (same value for all action dimensions), or a list (specific values for each dimension). If the
            latter, dimension should be the same as the control dimension for this controller

        input_min (float or Iterable of float): Minimum below which an inputted action will be clipped. Can be either be
            a scalar (same value for all action dimensions), or a list (specific values for each dimension). If the
            latter, dimension should be the same as the control dimension for this controller

        output_max (float or Iterable of float): Maximum which defines upper end of scaling range when scaling an input
            action. Can be either be a scalar (same value for all action dimensions), or a list (specific values for
            each dimension). If the latter, dimension should be the same as the control dimension for this controller

        output_min (float or Iterable of float): Minimum which defines upper end of scaling range when scaling an input
            action. Can be either be a scalar (same value for all action dimensions), or a list (specific values for
            each dimension). If the latter, dimension should be the same as the control dimension for this controller

        kp (float or Iterable of float): positional gain for determining desired torques based upon the pos / ori error.
            Can be either be a scalar (same value for all action dims), or a list (specific values for each dim)

        damping_ratio (float or Iterable of float): used in conjunction with kp to determine the velocity gain for
            determining desired torques based upon the joint pos errors. Can be either be a scalar (same value for all
            action dims), or a list (specific values for each dim)

        impedance_mode (str): Impedance mode with which to run this controller. Options are {"fixed", "variable",
            "variable_kp"}. If "fixed", the controller will have fixed kp and damping_ratio values as specified by the
            @kp and @damping_ratio arguments. If "variable", both kp and damping_ratio will now be part of the
            controller action space, resulting in a total action space of (6 or 3) + 6 * 2. If "variable_kp", only kp
            will become variable, with damping_ratio fixed at 1 (critically damped). The resulting action space will
            then be (6 or 3) + 6.

        kp_limits (2-list of float or 2-list of Iterable of floats): Only applicable if @impedance_mode is set to either
            "variable" or "variable_kp". This sets the corresponding min / max ranges of the controller action space
            for the varying kp values. Can be either be a 2-list (same min / max for all kp action dims), or a 2-list
            of list (specific min / max for each kp dim)

        damping_ratio_limits (2-list of float or 2-list of Iterable of floats): Only applicable if @impedance_mode is
            set to "variable". This sets the corresponding min / max ranges of the controller action space for the
            varying damping_ratio values. Can be either be a 2-list (same min / max for all damping_ratio action dims),
            or a 2-list of list (specific min / max for each damping_ratio dim)

        policy_freq (int): Frequency at which actions from the robot policy are fed into this controller

        position_limits (2-list of float or 2-list of Iterable of floats): Limits (m) below and above which the
            magnitude of a calculated goal eef position will be clipped. Can be either be a 2-list (same min/max value
            for all cartesian dims), or a 2-list of list (specific min/max values for each dim)

        orientation_limits (2-list of float or 2-list of Iterable of floats): Limits (rad) below and above which the
            magnitude of a calculated goal eef orientation will be clipped. Can be either be a 2-list
            (same min/max value for all joint dims), or a 2-list of list (specific min/mx values for each dim)

        interpolator_pos (Interpolator): Interpolator object to be used for interpolating from the current position to
            the goal position during each timestep between inputted actions

        interpolator_ori (Interpolator): Interpolator object to be used for interpolating from the current orientation
            to the goal orientation during each timestep between inputted actions

        control_ori (bool): Whether inputted actions will control both pos and ori or exclusively pos

        control_delta (bool): Whether to control the robot using delta or absolute commands (where absolute commands
            are taken in the world coordinate frame)

        uncouple_pos_ori (bool): Whether to decouple torques meant to control pos and torques meant to control ori

        **kwargs: Does nothing; placeholder to "sink" any additional arguments so that instantiating this controller
            via an argument dict that has additional extraneous arguments won't raise an error

    Raises:
        AssertionError: [Invalid impedance mode]
    """

    def __init__(
        self,
        sim,
        eef_name,
        joint_indexes,
        actuator_range,
        input_max=1,
        input_min=-1,
        output_max=(0.05, 0.05, 0.05, 0.5, 0.5, 0.5),
        output_min=(-0.05, -0.05, -0.05, -0.5, -0.5, -0.5),
        kp=150,
        damping_ratio=1,
        impedance_mode="fixed",
        kp_limits=(0, 300),
        damping_ratio_limits=(0, 100),
        policy_freq=20,
        position_limits=None,
        orientation_limits=None,
        interpolator_pos=None,
        interpolator_ori=None,
        control_ori=True,
        control_delta=True,
        uncouple_pos_ori=True,
        **kwargs,  # does nothing; used so no error raised when dict is passed with extra terms used previously
    ):

        super().__init__(
            sim,
            eef_name,
            joint_indexes,
            actuator_range,
        )
        # Determine whether this is pos ori or just pos
        self.use_ori = control_ori

        # Determine whether we want to use delta or absolute values as inputs
        self.use_delta = control_delta

        # Control dimension
        self.control_dim = 6 if self.use_ori else 3
        self.name_suffix = "POSE" if self.use_ori else "POSITION"

        # input and output max and min (allow for either explicit lists or single numbers)
        self.input_max = self.nums2array(input_max, self.control_dim)
        self.input_min = self.nums2array(input_min, self.control_dim)
        self.output_max = self.nums2array(output_max, self.control_dim)
        self.output_min = self.nums2array(output_min, self.control_dim)

        # kp kd
        self.kp = self.nums2array(kp, 6)
        self.kd = 2 * np.sqrt(self.kp) * damping_ratio

        # kp and kd limits
        self.kp_min = self.nums2array(kp_limits[0], 6)
        self.kp_max = self.nums2array(kp_limits[1], 6)
        self.damping_ratio_min = self.nums2array(damping_ratio_limits[0], 6)
        self.damping_ratio_max = self.nums2array(damping_ratio_limits[1], 6)

        # Verify the proposed impedance mode is supported
        assert impedance_mode in IMPEDANCE_MODES, (
            "Error: Tried to instantiate OSC controller for unsupported "
            "impedance mode! Inputted impedance mode: {}, Supported modes: {}".format(impedance_mode, IMPEDANCE_MODES)
        )

        # Impedance mode
        self.impedance_mode = impedance_mode

        # Add to control dim based on impedance_mode
        if self.impedance_mode == "variable":
            self.control_dim += 12
        elif self.impedance_mode == "variable_kp":
            self.control_dim += 6

        # limits
        self.position_limits = np.array(position_limits) if position_limits is not None else position_limits
        self.orientation_limits = np.array(orientation_limits) if orientation_limits is not None else orientation_limits

        # control frequency
        self.control_freq = policy_freq

        # interpolator
        self.interpolator_pos = interpolator_pos
        self.interpolator_ori = interpolator_ori

        # whether or not pos and ori want to be uncoupled
        self.uncoupling = uncouple_pos_ori

        # initialize goals based on initial pos / ori
        self.goal_ori = np.array(self.initial_ee_ori_mat)
        self.goal_pos = np.array(self.initial_ee_pos)

        self.relative_ori = np.zeros(3)
        self.ori_ref = None

    def set_goal(self, action, set_pos=None, set_ori=None):
        """
        Sets goal based on input @action. If self.impedance_mode is not "fixed", then the input will be parsed into the
        delta values to update the goal position / pose and the kp and/or damping_ratio values to be immediately updated
        internally before executing the proceeding control loop.

        Note that @action expected to be in the following format, based on impedance mode!

            :Mode `'fixed'`: [joint pos command]
            :Mode `'variable'`: [damping_ratio values, kp values, joint pos command]
            :Mode `'variable_kp'`: [kp values, joint pos command]

        Args:
            action (Iterable): Desired relative joint position goal state
            set_pos (Iterable): If set, overrides @action and sets the desired absolute eef position goal state
            set_ori (Iterable): IF set, overrides @action and sets the desired absolute eef orientation goal state
        """
        # Update state
        self.update()

        # Parse action based on the impedance mode, and update kp / kd as necessary
        if self.impedance_mode == "variable":
            damping_ratio, kp, delta = action[:6], action[6:12], action[12:]
            self.kp = np.clip(kp, self.kp_min, self.kp_max)
            self.kd = 2 * np.sqrt(self.kp) * np.clip(damping_ratio, self.damping_ratio_min, self.damping_ratio_max)
        elif self.impedance_mode == "variable_kp":
            kp, delta = action[:6], action[6:]
            self.kp = np.clip(kp, self.kp_min, self.kp_max)
            self.kd = 2 * np.sqrt(self.kp)  # critically damped
        else:  # This is case "fixed"
            delta = action

        # If we're using deltas, interpret actions as such
        if self.use_delta:
            if delta is not None:
                scaled_delta = self.scale_action(delta)
                if not self.use_ori and set_ori is None:
                    # Set default control for ori since user isn't actively controlling ori
                    set_ori = np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, -1.0]])
            else:
                scaled_delta = []
        # Else, interpret actions as absolute values
        else:
            if set_pos is None:
                set_pos = delta[:3]
            # Set default control for ori if we're only using position control
            if set_ori is None:
                set_ori = (
                    T.quat2mat(T.axisangle2quat(delta[3:6]))
                    if self.use_ori
                    else np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, -1.0]])
                )
            # No scaling of values since these are absolute values
            scaled_delta = delta

        # We only want to update goal orientation if there is a valid delta ori value OR if we're using absolute ori
        # use math.isclose instead of numpy because numpy is slow
        bools = [0.0 if math.isclose(elem, 0.0) else 1.0 for elem in scaled_delta[3:]]
        if sum(bools) > 0.0 or set_ori is not None:
            self.goal_ori = set_goal_orientation(
                scaled_delta[3:], self.ee_ori_mat, orientation_limit=self.orientation_limits, set_ori=set_ori
            )
        self.goal_pos = set_goal_position(
            scaled_delta[:3], self.ee_pos, position_limit=self.position_limits, set_pos=set_pos
        )

        if self.interpolator_pos is not None:
            self.interpolator_pos.set_goal(self.goal_pos)

        if self.interpolator_ori is not None:
            self.ori_ref = np.array(self.ee_ori_mat)  # reference is the current orientation at start
            self.interpolator_ori.set_goal(
                orientation_error(self.goal_ori, self.ori_ref)
            )  # goal is the total orientation error
            self.relative_ori = np.zeros(3)  # relative orientation always starts at 0

    def run_controller(self):
        """
        Calculates the torques required to reach the desired setpoint.

        Executes Operational Space Control (OSC) -- either position only or position and orientation.

        A detailed overview of derivation of OSC equations can be seen at:
        http://khatib.stanford.edu/publications/pdfs/Khatib_1987_RA.pdf

        Returns:
             np.array: Command torques
        """
        # Update state
        self.update()

        desired_pos = None
        # Only linear interpolator is currently supported
        if self.interpolator_pos is not None:
            # Linear case
            if self.interpolator_pos.order == 1:
                desired_pos = self.interpolator_pos.get_interpolated_goal()
            else:
                # Nonlinear case not currently supported
                pass
        else:
            desired_pos = np.array(self.goal_pos)

        if self.interpolator_ori is not None:
            # relative orientation based on difference between current ori and ref
            self.relative_ori = orientation_error(self.ee_ori_mat, self.ori_ref)

            ori_error = self.interpolator_ori.get_interpolated_goal()
        else:
            desired_ori = np.array(self.goal_ori)
            ori_error = orientation_error(desired_ori, self.ee_ori_mat)

        # Compute desired force and torque based on errors
        position_error = desired_pos - self.ee_pos
        vel_pos_error = -self.ee_pos_vel

        # F_r = kp * pos_err + kd * vel_err
        desired_force = np.multiply(np.array(position_error), np.array(self.kp[0:3])) + np.multiply(
            vel_pos_error, self.kd[0:3]
        )

        vel_ori_error = -self.ee_ori_vel

        # Tau_r = kp * ori_err + kd * vel_err
        desired_torque = np.multiply(np.array(ori_error), np.array(self.kp[3:6])) + np.multiply(
            vel_ori_error, self.kd[3:6]
        )

        # Compute nullspace matrix (I - Jbar * J) and lambda matrices ((J * M^-1 * J^T)^-1)
        lambda_full, lambda_pos, lambda_ori, nullspace_matrix = opspace_matrices(
            self.mass_matrix, self.J_full, self.J_pos, self.J_ori
        )

        # Decouples desired positional control from orientation control
        if self.uncoupling:
            decoupled_force = np.dot(lambda_pos, desired_force)
            decoupled_torque = np.dot(lambda_ori, desired_torque)
            decoupled_wrench = np.concatenate([decoupled_force, decoupled_torque])
        else:
            desired_wrench = np.concatenate([desired_force, desired_torque])
            decoupled_wrench = np.dot(lambda_full, desired_wrench)

        # Gamma (without null torques) = J^T * F + gravity compensations
        self.torques = np.dot(self.J_full.T, decoupled_wrench) + self.torque_compensation

        # Calculate and add nullspace torques (nullspace_matrix^T * Gamma_null) to final torques
        # Note: Gamma_null = desired nullspace pose torques, assumed to be positional joint control relative
        #                     to the initial joint positions
        self.torques += nullspace_torques(
            self.mass_matrix, nullspace_matrix, self.initial_joint, self.joint_pos, self.joint_vel
        )

        # Always run superclass call for any cleanups at the end
        super().run_controller()

        return self.torques

    def update_initial_joints(self, initial_joints):
        # First, update from the superclass method
        super().update_initial_joints(initial_joints)

        # We also need to reset the goal in case the old goals were set to the initial confguration
        self.reset_goal()

    def reset_goal(self):
        """
        Resets the goal to the current state of the robot
        """
        self.goal_ori = np.array(self.ee_ori_mat)
        self.goal_pos = np.array(self.ee_pos)

        # Also reset interpolators if required

        if self.interpolator_pos is not None:
            self.interpolator_pos.set_goal(self.goal_pos)

        if self.interpolator_ori is not None:
            self.ori_ref = np.array(self.ee_ori_mat)  # reference is the current orientation at start
            self.interpolator_ori.set_goal(
                orientation_error(self.goal_ori, self.ori_ref)
            )  # goal is the total orientation error
            self.relative_ori = np.zeros(3)  # relative orientation always starts at 0

    @property
    def control_limits(self):
        """
        Returns the limits over this controller's action space, overrides the superclass property
        Returns the following (generalized for both high and low limits), based on the impedance mode:

            :Mode `'fixed'`: [joint pos command]
            :Mode `'variable'`: [damping_ratio values, kp values, joint pos command]
            :Mode `'variable_kp'`: [kp values, joint pos command]

        Returns:
            2-tuple:

                - (np.array) minimum action values
                - (np.array) maximum action values
        """
        if self.impedance_mode == "variable":
            low = np.concatenate([self.damping_ratio_min, self.kp_min, self.input_min])
            high = np.concatenate([self.damping_ratio_max, self.kp_max, self.input_max])
        elif self.impedance_mode == "variable_kp":
            low = np.concatenate([self.kp_min, self.input_min])
            high = np.concatenate([self.kp_max, self.input_max])
        else:  # This is case "fixed"
            low, high = self.input_min, self.input_max
        return low, high

    @property
    def name(self):
        return "OSC_" + self.name_suffix



from os import path
from typing import Any, Dict, Optional, Tuple, Union, SupportsFloat

import numpy as np
from numpy.typing import NDArray

import gymnasium as gym
from gymnasium import error, spaces
from gymnasium.spaces import Space
from gymnasium.core import ObsType

import asyncio
import sys
from orca_gym import OrcaGymLocal
from orca_gym.protos.mjc_message_pb2_grpc import GrpcServiceStub 
from orca_gym.utils.rotations import mat2quat, quat2mat, quat_mul, quat2euler, euler2quat
from orca_gym.core.orca_gym_local import AnchorType, get_eq_type, CaptureMode

from orca_gym import OrcaGymModel
from orca_gym import OrcaGymData
from . import OrcaGymBaseEnv
from scipy.spatial.transform import Rotation as R

import grpc

from datetime import datetime
import time

class OrcaGymLocalEnv(OrcaGymBaseEnv):
    metadata = {'render_modes': ['human', 'none'], 'version': '0.0.1', 'render_fps': 30}
    
    def __init__(
        self,
        frame_skip: int,
        orcagym_addr: str,
        agent_names: list[str],
        time_step: float,        
        **kwargs        
    ):
        super().__init__(
            frame_skip = frame_skip,
            orcagym_addr = orcagym_addr,
            agent_names = agent_names,
            time_step = time_step,            
            **kwargs
        )

        render_fps = self.metadata.get("render_fps")

        # 用于异步渲染
        self._render_interval = 1.0 / render_fps
        self._render_time_step = time.perf_counter()

        # 用于同步渲染
        self._render_count_interval = self.realtime_step * render_fps
        self.render_count = 0
        self._last_frame_index = -1

        self.mj_forward()

        self._body_anchored = None
        self._anchor_body_name = "ActorManipulator_Anchor"
        self._anchor_dummy_body_name = "ActorManipulator_dummy"
        body_names = self.model.get_body_names()
        if (self._anchor_body_name in body_names and self._anchor_dummy_body_name in body_names):
            self._anchor_body_id = self.model.body_name2id(self._anchor_body_name)
            self._anchor_dummy_body_id = self.model.body_name2id(self._anchor_dummy_body_name)
        else:
            self._anchor_body_id = None
            self._anchor_dummy_body_id = None
            print(f"Warning: Anchor body {self._anchor_body_name} not found in the model. Actor manipulation is disabled.")

    def initialize_simulation(
        self,
    ) -> Tuple[OrcaGymModel, OrcaGymData]:
        print(f"Initializing simulation: Class: {self.__class__.__name__}")
        model_xml_path = self.loop.run_until_complete(self._load_model_xml())
        self.loop.run_until_complete(self._initialize_orca_sim(model_xml_path))
        model = self.gym.model
        data = self.gym.data
        return model, data
    
    async def _load_model_xml(self):
        model_xml_path = await self.gym.load_model_xml()
        return model_xml_path

    async def _initialize_orca_sim(self, model_xml_path):
        await self.gym.init_simulation(model_xml_path)
        return

    def initialize_grpc(self):
        self.channel = grpc.aio.insecure_channel(
            self.orcagym_addr,
            options=[
                ('grpc.max_receive_message_length', 1024 * 1024 * 1024),
                ('grpc.max_send_message_length', 1024 * 1024 * 1024),
            ]
        )
        self.stub = GrpcServiceStub(self.channel)
        self.gym = OrcaGymLocal(self.stub)
    
    def pause_simulation(self):
        self.loop.run_until_complete(self._pause_simulation())

    async def _pause_simulation(self):
        await self.gym.pause_simulation()

    async def _close_grpc(self):
        if self.channel:
            await self.channel.close()

    def close(self):
        self.loop.run_until_complete(self._close_grpc())

    async def _get_body_manipulation_anchored(self):
        return await self.gym.get_body_manipulation_anchored()
    
    def get_body_manipulation_anchored(self):
        return self.loop.run_until_complete(self._get_body_manipulation_anchored())
    
    async def _get_body_manipulation_movement(self):
        return await self.gym.get_body_manipulation_movement()
    
    def begin_save_video(self, file_path: str, capture_mode: CaptureMode = CaptureMode.ASYNC ):
        return self.loop.run_until_complete(self._begin_save_video(file_path, capture_mode=capture_mode))

    async def _begin_save_video(self, file_path, capture_mode: CaptureMode = CaptureMode.ASYNC ):
        return await self.gym.begin_save_video(file_path, capture_mode=capture_mode)

    def stop_save_video(self):
        return self.loop.run_until_complete(self._stop_save_video())

    async def _stop_save_video(self):
        return await self.gym.stop_save_video()

    def get_next_frame(self) -> int:
        current_frame = self.loop.run_until_complete(self._get_current_frame())
        if current_frame < 0:
            # 如果摄像头没有使能，会一直返回-1
            return current_frame
        
        for _ in range(10):
            if current_frame == self._last_frame_index:
                time.sleep(self.realtime_step)
            else:
                self._last_frame_index = current_frame
                return current_frame
        
        return current_frame
            

    def get_current_frame(self) -> int:
        return self.loop.run_until_complete(self._get_current_frame())

    async def _get_current_frame(self):
        return await self.gym.get_current_frame()

    def get_camera_time_stamp(self, last_frame_index) -> dict:
        return self.loop.run_until_complete(self._get_camera_time_stamp(last_frame_index))

    async def _get_camera_time_stamp(self, last_frame_index):
        return await self.gym.get_camera_time_stamp(last_frame_index)

    def get_frame_png(self, image_path):
        return self.loop.run_until_complete(self._get_frame_png(image_path))

    async def _get_frame_png(self, image_path):
        return await self.gym.get_frame_png(image_path)

    def get_body_manipulation_movement(self):
        actor_movement = self.loop.run_until_complete(self._get_body_manipulation_movement())
        delta_pos = actor_movement["delta_pos"]
        delta_quat = actor_movement["delta_quat"]
        return delta_pos, delta_quat

    def do_simulation(self, ctrl, n_frames) -> None:
        """
        Step the simulation n number of frames and applying a control action.
        """
        # Check control input is contained in the action space
        if np.array(ctrl).shape != (self.model.nu,):
            raise ValueError(
                f"Action dimension mismatch. Expected {(self.model.nu,)}, found {np.array(ctrl).shape}"
            )
        self._step_orca_sim_simulation(ctrl, n_frames)
        self.gym.update_data()


    @property
    def render_mode(self) -> str:
        if hasattr(self, "_render_mode"):
            return self._render_mode
        else:
            return "human"
        
    @property
    def is_subenv(self) -> bool:
        if hasattr(self, "_is_subenv"):
            return self._is_subenv
        else:
            return False

    @property
    def sync_render(self) -> bool:
        if hasattr(self, "_sync_render"):
            return self._sync_render
        else:
            return False

    def render(self):
        if self.render_mode not in ["human", "force"]:
            return

        if self.sync_render:
            self.render_count += self._render_count_interval
            if (self.render_count >= 1.0):
                self.loop.run_until_complete(self.gym.render())
                self.do_body_manipulation() # 只有在渲染时才处理锚点操作，否则也不会有场景视口交互行为
                self.render_count -= 1
        else:
            time_diff = time.perf_counter() - self._render_time_step
            if (time_diff > self._render_interval):
                self._render_time_step = time.perf_counter()
                self.loop.run_until_complete(self.gym.render())
                self.do_body_manipulation() # 只有在渲染时才处理锚点操作，否则也不会有场景视口交互行为

    def do_body_manipulation(self):
        if self._anchor_body_id is None:
            # 老版本不支持锚点操作
            return

        actor_anchored, anchor_type = self.get_body_manipulation_anchored()
        if actor_anchored is None:
            if self._body_anchored is not None:
                self.release_body_anchored()
            return
        
        if self._body_anchored is None:
            self.anchor_actor(actor_anchored, anchor_type)
        
        
        delta_pos, delta_quat = self.get_body_manipulation_movement()
        # print(f"Actor Manipulation: {actor_anchored}, Delta Pos: {delta_pos}, Delta Quat: {delta_quat}")
        delta_pos *= self._render_interval
        delta_euler = quat2euler(delta_quat)
        delta_euler *= self._render_interval
        delta_quat = euler2quat(delta_euler)

        # 移动和旋转锚点
        anchor_xpos, anchor_xmat, anchor_xquat = self.get_body_xpos_xmat_xquat([self._anchor_body_name])
        if anchor_xpos is None or anchor_xmat is None or anchor_xquat is None:
            print(f"Warning: Anchor body {self._anchor_body_name} not found in the simulation. Cannot anchor.")
            return

        # 更新锚点位置
        anchor_xpos = anchor_xpos + delta_pos
        # 更新锚点四元数
        anchor_xquat = quat_mul(
            anchor_xquat,
            delta_quat
        )

        # 更新锚点的位置和四元数
        self.set_mocap_pos_and_quat({
            self._anchor_body_name: {
                "pos": anchor_xpos,
                "quat": anchor_xquat
            }
        })
        self.mj_forward()

        # print(f"Updated anchor position: {anchor_xpos}, quaternion: {anchor_xquat}")


    def release_body_anchored(self):
        if self._body_anchored is not None:
            self.update_anchor_equality_constraints(self._anchor_dummy_body_name, AnchorType.NONE)

            self.set_mocap_pos_and_quat({
                self._anchor_body_name: {
                    "pos": np.array([0.0, 0.0, -1000.0], dtype=np.float64),
                    "quat": np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)  # Reset to a far position
                }
            })
            self.mj_forward()

            # print(f"Released actor: {self._body_anchored}")
            self._body_anchored = None
        else:
            print("No actor is currently anchored.")

    def anchor_actor(self, actor_name: str, anchor_type: AnchorType):
        assert self._body_anchored is None, "An actor is already anchored. Please release it first."
        
        # 获取actor的位姿和四元数
        actor_xpos, actor_xmat, actor_xquat = self.get_body_xpos_xmat_xquat([actor_name])
        if actor_xpos is None or actor_xmat is None or actor_xquat is None:
            print(f"Warning: Actor {actor_name} not found in the simulation. Cannot anchor.")
            return
        
        # 将锚点位置设置为actor的位姿

        mocap_pos_and_quat_dict = {
            self._anchor_body_name: {
                "pos": actor_xpos,
                "quat": actor_xquat if anchor_type == AnchorType.WELD else np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)
            }
        }
        self.set_mocap_pos_and_quat(mocap_pos_and_quat_dict)
        self.mj_forward()

        self.update_anchor_equality_constraints(actor_name, anchor_type)


        self._body_anchored = actor_name
        # print(f"Anchored actor: {self._body_anchored} at position {actor_xpos} with quaternion {actor_xquat}")

    def update_anchor_equality_constraints(self, actor_name: str, anchor_type: AnchorType):
        # 更新锚点的平衡约束
        eq_list = self.model.get_eq_list()
        if eq_list is None or len(eq_list) == 0:
            raise ValueError("No equality constraints found in the model.")
        
        for eq in eq_list:
            old_obj1_id = eq["obj1_id"]
            old_obj2_id = eq["obj2_id"]
            if eq["obj1_id"] == self._anchor_body_id:
                eq["obj2_id"] = self.model.body_name2id(actor_name)
                self.gym.modify_equality_objects(
                    old_obj1_id= old_obj1_id,
                    old_obj2_id= old_obj2_id,
                    new_obj1_id= eq["obj1_id"],
                    new_obj2_id= eq["obj2_id"]
                )
                eq["eq_type"] = get_eq_type(anchor_type)
                # print(f"Anchoring actor {actor_name} to anchor body {self._anchor_body_name}")
                break
            elif eq["obj2_id"] == self._anchor_body_id:
                eq["obj1_id"] = self.model.body_name2id(actor_name)
                self.gym.modify_equality_objects(
                    old_obj1_id= old_obj1_id,
                    old_obj2_id= old_obj2_id,
                    new_obj1_id= eq["obj1_id"],
                    new_obj2_id= eq["obj2_id"]
                )
                eq["eq_type"] = get_eq_type(anchor_type)
                # print(f"Anchoring anchor body {self._anchor_body_name} to actor {actor_name}")
                break

        self.gym.update_equality_constraints(eq_list)
        self.mj_forward()

    def set_ctrl(self, ctrl):
        self.gym.set_ctrl(ctrl)

    def mj_step(self, nstep):
        self.gym.mj_step(nstep)

    def mj_forward(self):
        self.gym.mj_forward()

    def mj_jacBody(self, jacp, jacr, body_id):
        self.gym.mj_jacBody(jacp, jacr, body_id)

    def mj_jacSite(self, jacp, jacr, site_name):
        self.gym.mj_jacSite(jacp, jacr, site_name)

    def _step_orca_sim_simulation(self, ctrl, n_frames):
        self.set_ctrl(ctrl)
        self.mj_step(nstep=n_frames)

    def set_time_step(self, time_step):
        self.time_step = time_step
        self.realtime_step = time_step * self.frame_skip
        self.gym.set_time_step(time_step)
        self.loop.run_until_complete(self.gym.set_timestep_remote(time_step))
        return

    def update_data(self):
        self.gym.update_data()
        return

    def reset_simulation(self):
        self.gym.load_initial_frame()
        self.gym.update_data()

    def init_qpos_qvel(self):
        self.gym.update_data()
        self.init_qpos = self.gym.data.qpos.ravel().copy()
        self.init_qvel = self.gym.data.qvel.ravel().copy()

    def query_joint_offsets(self, joint_names) -> Tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
        qpos_offsets, qvel_offsets, qacc_offsets = self.gym.query_joint_offsets(joint_names)
        return qpos_offsets, qvel_offsets, qacc_offsets
    
    def query_joint_lengths(self, joint_names):
        qpos_lengths, qvel_lengths, qacc_lengths = self.gym.query_joint_lengths(joint_names)
        return qpos_lengths, qvel_lengths, qacc_lengths
    
    def get_body_xpos_xmat_xquat(self, body_name_list):
        body_dict = self.gym.query_body_xpos_xmat_xquat(body_name_list)
        if len(body_dict) != len(body_name_list):
            print("Body Nmae List: ", body_name_list)
            print("Body Dict: ", body_dict)
            raise ValueError("Some body names are not found in the simulation.")
        xpos = np.array([body_dict[body_name]['Pos'] for body_name in body_name_list]).flat.copy()
        xmat = np.array([body_dict[body_name]['Mat'] for body_name in body_name_list]).flat.copy()
        xquat = np.array([body_dict[body_name]['Quat'] for body_name in body_name_list]).flat.copy()
        return xpos, xmat, xquat
    
    def query_sensor_data(self, sensor_names):
        sensor_data_dict = self.gym.query_sensor_data(sensor_names)
        return sensor_data_dict
    
    def query_joint_qpos(self, joint_names):
        joint_qpos_dict = self.gym.query_joint_qpos(joint_names)
        return joint_qpos_dict
    
    def query_joint_qvel(self, joint_names):
        joint_qvel_dict = self.gym.query_joint_qvel(joint_names)
        return joint_qvel_dict
    
    def query_joint_qacc(self, joint_names):
        joint_qacc_dict = self.gym.query_joint_qacc(joint_names)
        return joint_qacc_dict
    
    def jnt_qposadr(self, joint_name):
        joint_qposadr = self.gym.jnt_qposadr(joint_name)
        return joint_qposadr
    
    def jnt_dofadr(self, joint_name):
        joint_dofadr = self.gym.jnt_dofadr(joint_name)
        return joint_dofadr
        
    def query_site_pos_and_mat(self, site_names):
        query_dict = self.gym.query_site_pos_and_mat(site_names)
        site_dict = {}
        for site in query_dict:
            site_dict[site] = {
                'xpos': np.array(query_dict[site]['xpos']),
                'xmat': np.array(query_dict[site]['xmat'])
            }
        return site_dict
    
    def query_site_pos_and_quat(self, site_names) -> Dict[str, Dict[str, Union[NDArray[np.float64], NDArray[np.float64]]]]:
        query_dict = self.gym.query_site_pos_and_mat(site_names)
        site_dict = {}
        for site in query_dict:
            site_dict[site] = {
                'xpos': np.array(query_dict[site]['xpos']),
                'xquat': mat2quat(np.array(query_dict[site]['xmat']).reshape(3, 3))
            }
        return site_dict
    
    def query_site_size(self, site_names):
        site_size_dict = self.gym.query_site_size(site_names)
        return site_size_dict


    def query_site_pos_and_quat_B(self, site_names, base_body_list) -> Dict[str, Dict[str, Union[NDArray[np.float32], NDArray[np.float32]]]]:
        site_dict = self.gym.query_site_pos_and_mat(site_names)
        base_pos, _, base_quat = self.get_body_xpos_xmat_xquat(base_body_list)
        site_pos_quat_B = {}
        for site_name, site_value in site_dict.items():
            ee_pos = np.array(site_value['xpos'])
            ee_quat = mat2quat(np.array(site_value['xmat']).reshape(3, 3))

            # 转换为SciPy需要的[x,y,z,w]格式
            rot_base = R.from_quat([base_quat[1], base_quat[2], base_quat[3], base_quat[0]])
            rot_base_inv = rot_base.inv()

            rot_ee = R.from_quat([ee_quat[1], ee_quat[2], ee_quat[3], ee_quat[0]])
            relative_rot_ee = rot_base_inv * rot_ee
            relative_pos_ee = rot_base_inv.apply(ee_pos - base_pos)

            site_pos_quat_B[site_name] = {}
            site_pos_quat_B[site_name]["xpos"] = relative_pos_ee
            site_pos_quat_B[site_name]["xquat"] = relative_rot_ee.as_quat().astype(np.float32)

        return site_pos_quat_B


    def set_joint_qpos(self, joint_qpos):
        self.gym.set_joint_qpos(joint_qpos)

    def set_joint_qvel(self, joint_qvel):
        self.gym.set_joint_qvel(joint_qvel)
    
    def query_site_xvalp_xvalr(self, site_names) -> Tuple[Dict[str, NDArray[np.float64]], Dict[str, NDArray[np.float64]]]:
        query_dict = self.gym.mj_jac_site(site_names)
        xvalp_dict = {}
        xvalr_dict = {}
        for site in query_dict:
            xvalp_dict[site] = np.array(query_dict[site]['jacp']).reshape(3, -1) @ self.data.qvel
            xvalr_dict[site] = np.array(query_dict[site]['jacr']).reshape(3, -1) @ self.data.qvel

        return xvalp_dict, xvalr_dict        
    
    def query_site_xvalp_xvalr_B(self, site_names, base_body_list) -> Tuple[Dict[str, NDArray[np.float32]], Dict[str, NDArray[np.float32]]]:
        query_dict = self.gym.mj_jac_site(site_names)
        _, base_mat, _ = self.get_body_xpos_xmat_xquat(base_body_list)
        xvalp_dict = {}
        xvalr_dict = {}
        for site in query_dict:
            ee_xvalp = np.array(query_dict[site]['jacp']).reshape(3, -1) @ self.data.qvel
            ee_xvalr = np.array(query_dict[site]['jacr']).reshape(3, -1) @ self.data.qvel

            # 目前只有固定基座，不涉及基座的速度
            base_xvalp = np.zeros(3)
            base_xvalr = np.zeros(3)

            base_mat = base_mat.reshape(3, 3)
            linear_vel_B = base_mat.T @ (ee_xvalp - base_xvalp)
            angular_vel_B = base_mat.T @ (ee_xvalr - base_xvalr)

            xvalp_dict[site] = linear_vel_B.astype(np.float32)
            xvalr_dict[site] = angular_vel_B.astype(np.float32)

        return xvalp_dict, xvalr_dict
    
    def update_equality_constraints(self, eq_list):
        self.gym.update_equality_constraints(eq_list)

    def set_mocap_pos_and_quat(self, mocap_pos_and_quat_dict):
        send_remote = self.render_mode == "human" and not self.is_subenv
        self.loop.run_until_complete(self.gym.set_mocap_pos_and_quat(mocap_pos_and_quat_dict, send_remote))

    def query_contact_simple(self):
        return self.gym.query_contact_simple()
    
    def set_geom_friction(self, geom_friction_dict):
        self.gym.set_geom_friction(geom_friction_dict)

    def add_extra_weight(self, weight_load_dict):
        self.gym.add_extra_weight(weight_load_dict)
    
    def query_contact_force(self, contact_ids):
        contact_force = self.gym.query_contact_force(contact_ids)
        return contact_force
    
    def get_cfrc_ext(self):
        cfrc_ext = self.gym.get_cfrc_ext()
        return cfrc_ext

    def query_actuator_torques(self, actuator_names):
        actuator_torques = self.gym.query_actuator_torques(actuator_names)
        return actuator_torques

    def query_joint_dofadrs(self, joint_names):
        joint_dofadrs = self.gym.query_joint_dofadrs(joint_names)
        return joint_dofadrs
        
    def get_goal_bounding_box(self, geom_name):
        geom_size = self.gym.get_goal_bounding_box(geom_name)
        return geom_size

    def query_velocity_body_B(self, ee_body, base_body):
        return self.gym.query_velocity_body_B(ee_body, base_body)

    def query_position_body_B(self, ee_body, base_body):
        position_body_B = self.gym.query_position_body_B(ee_body, base_body)
        return position_body_B

    def query_orientation_body_B(self, ee_body, base_body):
        orientation_body_B = self.gym.query_orientation_body_B(ee_body, base_body)
        return orientation_body_B

    def query_joint_axes_B(self, joint_names, base_body):
        joint_axes_B = self.gym.query_joint_axes_B(joint_names, base_body)
        return joint_axes_B

    def query_robot_velocity_odom(self, base_body, initial_base_pos, initial_base_quat):
        linear, angular = self.gym.query_robot_velocity_odom(base_body, initial_base_pos, initial_base_quat)
        return linear, angular

    def query_robot_position_odom(self, base_body, initial_base_pos, initial_base_quat):
        robot_position_odom = self.gym.query_robot_position_odom(base_body, initial_base_pos, initial_base_quat)
        return robot_position_odom

    def query_robot_orientation_odom(self, base_body, initial_base_pos, initial_base_quat):
        robot_orientation_odom = self.gym.query_robot_orientation_odom(base_body, initial_base_pos, initial_base_quat)
        return robot_orientation_odom
    
    def set_actuator_trnid(self, actuator_id, trnid):
        self.gym.set_actuator_trnid(actuator_id, trnid)
        return

    async def _load_content_file(self, content_file_name, remote_file_dir="", local_file_dir="", temp_file_path=None):
        content_file_path = await self.gym.load_content_file(content_file_name, remote_file_dir, local_file_dir, temp_file_path)
        return content_file_path

    def load_content_file(self, content_file_name, remote_file_dir="", local_file_dir="", temp_file_path=None):
        return self.loop.run_until_complete(self._load_content_file(content_file_name, remote_file_dir, local_file_dir, temp_file_path))



from ast import main
import h5py
import json
import numpy as np
import logging as log
import glob
import os
import shutil
from datetime import datetime

from orca_gym.adapters.robomimic.dataset_util import DatasetReader, DatasetWriter


def _check_dataset(dataset_file, verbose=False) -> bool:

    # extract demonstration list from file
    all_filter_keys = None
    f = h5py.File(dataset_file, "r")

    # use all demonstrations
    demos = sorted(list(f["data"].keys()))
    
    if len(demos) == 0:
        print("")
        print("Dataset {} has no demonstrations.".format(dataset_file))
        return True

    # extract filter key information
    if "mask" in f:
        all_filter_keys = {}
        for fk in f["mask"]:
            fk_demos = sorted([elem.decode("utf-8") for elem in np.array(f["mask/{}".format(fk)])])
            all_filter_keys[fk] = fk_demos

    # put demonstration list in increasing episode order
    inds = np.argsort([int(elem[5:]) for elem in demos])
    demos = [demos[i] for i in inds]

    # extract length of each trajectory in the file
    traj_lengths = []
    action_min = np.inf
    action_max = -np.inf
    for ep in demos:
        traj_lengths.append(f["data/{}/actions".format(ep)].shape[0])
        action_min = min(action_min, np.min(f["data/{}/actions".format(ep)][()]))
        action_max = max(action_max, np.max(f["data/{}/actions".format(ep)][()]))
    traj_lengths = np.array(traj_lengths)

    # report statistics on the data
    print("")
    print("total transitions: {}".format(np.sum(traj_lengths)))
    print("total trajectories: {}".format(traj_lengths.shape[0]))
    print("traj length mean: {}".format(np.mean(traj_lengths)))
    print("traj length std: {}".format(np.std(traj_lengths)))
    print("traj length min: {}".format(np.min(traj_lengths)))
    print("traj length max: {}".format(np.max(traj_lengths)))
    print("action min: {}".format(action_min))
    print("action max: {}".format(action_max))
    print("")
    print("==== Filter Keys ====")
    if all_filter_keys is not None:
        for fk in all_filter_keys:
            print("filter key {} with {} demos".format(fk, len(all_filter_keys[fk])))
    else:
        print("no filter keys")
    print("")
    if verbose:
        if all_filter_keys is not None:
            print("==== Filter Key Contents ====")
            for fk in all_filter_keys:
                print("filter_key {} with {} demos: {}".format(fk, len(all_filter_keys[fk]), all_filter_keys[fk]))
        print("")
    env_meta = json.loads(f["data"].attrs["env_args"])
    print("==== Env Meta ====")
    print(json.dumps(env_meta, indent=4))
    print("")

    print("==== Dataset Structure ====")
    for ep in demos:
        print("episode {} with {} transitions".format(ep, f["data/{}".format(ep)].attrs["num_samples"]))
        for k in f["data/{}".format(ep)]:
            if k in ["obs", "next_obs"]:
                print("    key: {}".format(k))
                for obs_k in f["data/{}/{}".format(ep, k)]:
                    shape = f["data/{}/{}/{}".format(ep, k, obs_k)].shape
                    print("        observation key {} with shape {}".format(obs_k, shape))
            elif isinstance(f["data/{}/{}".format(ep, k)], h5py.Dataset):
                key_shape = f["data/{}/{}".format(ep, k)].shape
                print("    key: {} with shape {}".format(k, key_shape))

        if not verbose:
            break

    f.close()

    # maybe display error message
    print("")
    if (action_min < -1.) or (action_max > 1.):
        log.warning("Dataset should have actions in [-1., 1.] but got bounds [{}, {}]".format(action_min, action_max))
        return False
    else:
        print("Checking dataset {} complete. Dataset is valid.".format(dataset_file))
        
    return True

def process_check(dataset_files, verbose=False):
    check_successed = True
    checked_files_count = 0
    for dataset in dataset_files:
        if _check_dataset(dataset, verbose=verbose) == False:
            print("Dataset {} is invalid.".format(dataset))
            check_successed = False
            break
        checked_files_count += 1
        
    if check_successed:
        print("")
        print("All datasets are valid. Checked {} files.".format(checked_files_count))

def _check_combination(dataset_files):
    env_names = []
    env_versions = []
    env_kwargs = []
    for dataset in dataset_files:
        reader = DatasetReader(dataset)
        env_names.append(reader.get_env_name())
        env_versions.append(reader.get_env_version())
        env_kwargs.append(reader.get_env_kwargs())
        
    if len(set(env_names)) > 1:
        print("Different env names in datasets: {}".format(set(env_names)))
        return False
    if len(set(env_versions)) > 1:
        print("Different env versions in datasets: {}".format(set(env_versions)))
        return False
    if not all(env_kwargs_item == env_kwargs[0] for env_kwargs_item in env_kwargs):
        print(f"Different env kwargs in datasets: {env_kwargs}")
        return False
    
    return True

def get_dataset_prefix(dataset_file):
    reader = DatasetReader(dataset_file)
    env_name = reader.get_env_name()
    env_name = env_name.split("-OrcaGym-")[0]
    env_version = reader.get_env_version()
    env_kwargs = reader.get_env_kwargs()
    if "task" not in env_kwargs:
        env_kwargs["task"] = "default_task"
    env_task = env_kwargs["task"]
    prefix = f"{env_name}_{env_version}_{env_task}"
    prefix = prefix.replace(" ", "_")
    prefix = prefix.replace(".", "_")
    return prefix

def process_combine(dataset_files, output_file):
    if not _check_combination(dataset_files):
        print("Datasets are not compatible for combination. Cannot combine datasets.")
        exit(1)
        
    # combine datasets
    writer = None
    for dataset in dataset_files:
        reader = DatasetReader(dataset)
        if writer is None:
            env_name = reader.get_env_name()
            env_version = reader.get_env_version()
            env_kwargs = reader.get_env_kwargs()
            writer = DatasetWriter(output_file, env_name, env_version, env_kwargs, specific_file_path=output_file)
            
        demo_names = reader.get_demo_names()
        for demo_name in demo_names:
            demo_data = reader.get_demo_data(demo_name)
            writer.add_demo_data(demo_data)
    
    writer.shuffle_demos()
    writer.finalize()
    print("Combined datasets written to {}".format(output_file))

def glob_dataset_filenames(dataset_files):
    matched_files = []
    for pattern in dataset_files:
        # 构建两种递归模式：匹配任意子目录的.h5和.hdf5文件
        patterns = [
            os.path.join(pattern, '**', '*.h5'),    # 递归匹配.h5
            os.path.join(pattern, '**', '*.hdf5'),  # 递归匹配.hdf5
        ]
        for p in patterns:
            # 使用递归通配符并确保区分大小写
            matched_files.extend(glob.glob(p, recursive=True))
    return matched_files

def process_update_kwargs(dataset_files, kwargs):
    for dataset in dataset_files:
        reader = DatasetReader(dataset)
        env_name = reader.get_env_name()
        env_version = reader.get_env_version()
        env_kwargs = reader.get_env_kwargs()
        
        for key in kwargs:
            if key not in env_kwargs:
                print(f"Key {key} not in env_kwargs. Skip updating.")
                continue
            env_kwargs[key] = kwargs[key]
            print(f"Update key {key} to {kwargs[key]}")
        
        
        writer = DatasetWriter(dataset, env_name, env_version, env_kwargs)
        writer.set_env_kwargs(env_kwargs)
        writer.finalize()
def split_demos(dataset_file, output_dir, custom_name="demo"):
    """
    将一个 HDF5 文件中的所有 demonstrations 拆分为单独的 HDF5 文件，
    并自动重命名输出目录为：
    自定义名字_原始文件大小GB_文件数量_所有obs/camera_head视频时长总和秒
    """
    original_file_size = os.path.getsize(dataset_file)
    original_file_size_gb = f"{original_file_size/1024/1024/1024:.2f}GB"

    temp_dir = output_dir + "_temp"
    os.makedirs(temp_dir, exist_ok=True)

    with h5py.File(dataset_file, "r") as f:
        demos = sorted(list(f["data"].keys()))

        if len(demos) == 0:
            print(f"Dataset {dataset_file} has no demonstrations.")
            return

        # for i, demo in enumerate(demos):
        #     output_file = os.path.join(temp_dir, f"{demo}.hdf5")
            
        #     with h5py.File(output_file, "w") as out_f:
        #         out_data = out_f.create_group("data")
        #         demo_group = f["data"][demo]
                
        #         def copy_all(name, obj):
        #             if isinstance(obj, h5py.Dataset):
        #                 out_data.create_dataset(name, data=obj[()],)
        #             else:
        #                 print(f"Skipping non-dataset key: {name}")
                
        #         demo_group.visititems(copy_all)
        #         for attr_key in demo_group.attrs.keys():
        #             out_data.attrs[attr_key] = demo_group.attrs[attr_key]

        #     print(f"Saved demonstration {demo} to {output_file}")
        for i, demo in enumerate(demos):
            output_file = os.path.join(temp_dir, f"{demo}.hdf5")
            
            with h5py.File(output_file, "w") as out_f:
                out_data = out_f.create_group("data")
                demo_group = f["data"][demo]
                
                # 1. 先创建所有必要的组结构
                if "obs" in demo_group:
                    out_data.create_group('obs')
                if "next_obs" in demo_group:
                    out_data.create_group('next_obs')
                
                # 2. 显式复制obs数据并压缩
                if "obs" in demo_group:
                    for obs_key in demo_group["obs"].keys():
                        obs_data = demo_group["obs"][obs_key]
                        out_data["obs"].create_dataset(obs_key, data=obs_data, compression="gzip", compression_opts=4)
                
                # 3. 显式复制next_obs数据并压缩
                if "next_obs" in demo_group:
                    for next_obs_key in demo_group["next_obs"].keys():
                        next_obs_data = demo_group["next_obs"][next_obs_key]
                        out_data["next_obs"].create_dataset(next_obs_key, data=next_obs_data, compression="gzip", compression_opts=4)
                
                # 4. 使用visititems复制其他数据集
                def copy_other_datasets(name, obj):
                    if isinstance(obj, h5py.Dataset):
                        # 跳过已经处理过的obs和next_obs数据集
                        if name.startswith("obs/") or name.startswith("next_obs/"):
                            return
                        out_data.create_dataset(name, data=obj[()])
                    else:
                        print(f"Skipping non-dataset key: {name}")
                
                demo_group.visititems(copy_other_datasets)
                
                # 5. 复制所有属性
                for attr_key in demo_group.attrs.keys():
                    out_data.attrs[attr_key] = demo_group.attrs[attr_key]

            print(f"Saved demonstration {demo} to {output_file}")

    print(f"All demonstrations from {dataset_file} have been saved to temporary directory {temp_dir}")

    total_files = len(demos)
    total_video_duration_s = 0.0
   
    if os.path.exists(temp_dir):
        for filename in os.listdir(temp_dir):
            if filename.endswith(".hdf5"):
                filepath = os.path.join(temp_dir, filename)
                with h5py.File(filepath, "r") as f:
                    if "data" in f and "obs" in f["data"] and "camera_head" in f["data"]["obs"]:
                        dataset = f["data"]["obs"]["camera_head"]
                        fps = 30.0
                        video_duration_s = dataset.shape[0] / fps
                        total_video_duration_s += video_duration_s

    timestamp_str = "_".join(os.path.basename(dataset_file).split("_")[-2:])
    final_dir_name = f"{custom_name}_{original_file_size_gb}_{total_files}_{total_video_duration_s:.2f}s"
    final_dir = os.path.join(os.path.dirname(output_dir), final_dir_name)
    counter = 1
    while os.path.exists(final_dir):
        final_dir = os.path.join(os.path.dirname(output_dir), f"{final_dir_name}_{counter}")
        counter += 1
    shutil.move(temp_dir, final_dir)
    print(f"Renamed directory to: {final_dir}")

def merge_splits(base_dir, custom_name="demo"):
    """
    合并相同custom_name的所有拆分结果，生成一个总目录
    格式与之前相同：自定义名字_原始文件大小GB_文件数量_所有obs/camera_head视频时长总和秒
    """
    split_dirs = []
    for dirname in os.listdir(base_dir):
        if dirname.startswith(custom_name + "_") and not dirname.endswith("_merged_"):
            split_dirs.append(os.path.join(base_dir, dirname))

    if not split_dirs:
        print(f"No split directories found for {custom_name}")
        return
    accumulated_size_gb = 0.0
    accumulated_files = 0
    accumulated_video_duration_s = 0.0

    for split_dir in split_dirs:
        metadata_file = os.path.join(split_dir, "metadata.json")
        if os.path.exists(metadata_file):
            with open(metadata_file, "r") as f:
                metadata = json.load(f)
                accumulated_size_gb += metadata["original_file_size_gb"]
                accumulated_files += metadata["total_files"]
        else:
            dir_basename = os.path.basename(split_dir)
            parts = [p for p in dir_basename.split("_") if p]
            
            try:
                name_index = parts.index(custom_name)
                if name_index + 2 >= len(parts):
                    print(f"Warning: Directory name {dir_basename} has insufficient parts, skipping")
                    continue
                
                size_part = parts[name_index+1]
                size_gb = float(size_part[:-2])
                
                files_part = parts[name_index+2]
                files = int(files_part)
                
                accumulated_size_gb += size_gb
                accumulated_files += files
            except (ValueError, IndexError) as e:
                print(f"Warning: Could not parse directory name {dir_basename}, skipping. Error: {str(e)}")
                continue
        if os.path.exists(split_dir):
            for filename in os.listdir(split_dir):
                if filename.endswith(".hdf5"):
                    filepath = os.path.join(split_dir, filename)
                    with h5py.File(filepath, "r") as f:
                        if "data" in f and "obs" in f["data"] and "camera_head" in f["data"]["obs"]:
                            dataset = f["data"]["obs"]["camera_head"]
                            fps = 30.0
                            video_duration_s = dataset.shape[0] / fps
                            accumulated_video_duration_s += video_duration_s
    timestamp_str = datetime.now().strftime("%Y%m%d_%H%M%S")
    merged_dir_name = f"{custom_name}_{accumulated_size_gb:.2f}GB_{accumulated_files}_files_{accumulated_video_duration_s:.2f}s"
    merged_dir = os.path.join(base_dir, merged_dir_name)
    os.makedirs(merged_dir, exist_ok=True)

    demo_file_counter = {}

    for split_dir in split_dirs:
        for filename in os.listdir(split_dir):
            src = os.path.join(split_dir, filename)
            
            if os.path.isdir(src):
                dst = os.path.join(merged_dir, filename)
                shutil.copytree(src, dst, dirs_exist_ok=True)
            else:
                if filename.endswith(".hdf5"):
                    if filename in demo_file_counter:
                        base_name = filename[:-5]
                        ext = ".hdf5"
                        new_filename = f"{base_name}_{demo_file_counter[filename]}{ext}"
                        demo_file_counter[filename] += 1
                        dst = os.path.join(merged_dir, new_filename)
                    else:
                        demo_file_counter[filename] = 1
                        dst = os.path.join(merged_dir, filename)

                    shutil.copy2(src, dst)
                else:
                    dst = os.path.join(merged_dir, filename)
                    shutil.copy2(src, dst)
    metadata = {
        "original_file_size_gb": round(accumulated_size_gb, 2),
        "total_files": accumulated_files,
        "total_video_duration_s": round(accumulated_video_duration_s, 2),
        "timestamp": timestamp_str
    }
    with open(os.path.join(merged_dir, "metadata.json"), "w") as f:
        json.dump(metadata, f, indent=2)

    print(f"Merged {len(split_dirs)} splits into: {merged_dir}")
    
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Dataset processing utility")
    subparsers = parser.add_subparsers(dest="command", required=True)

    # === check command ===
    parser_check = subparsers.add_parser("check", help="Check dataset validity")
    parser_check.add_argument("datasets", nargs="+", help="Dataset file(s) to check")
    parser_check.add_argument("-v", "--verbose", action="store_true", help="Verbose output")

    # === combine command ===
    parser_combine = subparsers.add_parser("combine", help="Combine multiple datasets")
    parser_combine.add_argument("datasets", nargs="+", help="Dataset files to combine")
    parser_combine.add_argument("-o", "--output", required=True, help="Output file for combined dataset")

    # === update-kwargs command ===
    parser_update = subparsers.add_parser("update-kwargs", help="Update environment kwargs")
    parser_update.add_argument("datasets", nargs="+", help="Dataset files to update")
    parser_update.add_argument("--set", nargs=2, action="append", metavar=("KEY", "VALUE"),
                               help="Set key to value in env_kwargs")
     # === split command ===
    parser_split = subparsers.add_parser("split", help="Split a dataset into individual demonstrations")
    parser_split.add_argument("dataset", help="Dataset file to split (relative to the base directory)")
    parser_split.add_argument("-o", "--output_dir", required=True, help="Output directory name (relative to the base directory)")
    parser_split.add_argument("--name", default="demo", help="Custom name prefix for output")

    # === merge command ===
    parser_merge = subparsers.add_parser("merge", help="Merge split directories")
    parser_merge.add_argument("output_dir", help="Base output directory containing splits")
    parser_merge.add_argument("--name", default="demo", help="Custom name prefix of splits to merge")
    args = parser.parse_args()

    # 定义输入和输出的基目录
    # BASE_INPUT_DIR = "/home/orcatest/Orcagym_kps/OrcaGym/examples/openpi/records_tmp/shop/"
    # BASE_OUTPUT_DIR = "/home/orcatest/Orcagym_kps/OrcaGym/examples/openpi/records_tmp/shop/"
    SCRIPT_PATH = os.path.abspath(__file__)
    # 向上递归查找 OrcaGym 根目录（假设脚本在 orca_gym/scripts 下，根目录是 OrcaGym 的父目录）
    ORCA_GYM_ROOT = os.path.dirname(os.path.dirname(os.path.dirname(SCRIPT_PATH)))

    # 拼接绝对路径（最终输入输出目录）
    BASE_INPUT_DIR = os.path.join(ORCA_GYM_ROOT)
    BASE_OUTPUT_DIR = os.path.join(ORCA_GYM_ROOT)

    if args.command == "check":
        files = glob_dataset_filenames(args.datasets)
        process_check(files, verbose=args.verbose)

    elif args.command == "combine":
        files = glob_dataset_filenames(args.datasets)
        process_combine(files, output_file=args.output)

    elif args.command == "update-kwargs":
        files = glob_dataset_filenames(args.datasets)
        if args.set:
            kwargs = {k: json.loads(v) for k, v in args.set}
            process_update_kwargs(files, kwargs)
        else:
            print("No key-value pairs provided for update.")
    elif args.command == "split":
        input_file = os.path.join(BASE_INPUT_DIR, args.dataset)
        output_dir = os.path.join(BASE_OUTPUT_DIR, args.output_dir)
        split_demos(input_file, output_dir, args.name)

    elif args.command == "merge":
        merge_splits(args.output_dir, args.name)


import json
import os

# 打开并读取 JSON 文件
def load_config(config_name) -> dict:
    # 获取当前路径
    current_path = os.path.dirname(__file__)
    print("current_path: ", current_path)
    # 获取配置文件路径
    config_file_path = current_path + f"/config/{config_name}.json"

    with open(config_file_path, 'r') as file:
        config = json.load(file)
    return config


from ray.rllib.core.rl_module.rl_module import RLModule, RLModuleConfig
from ray.rllib.core.rl_module.torch.torch_rl_module import TorchRLModule
from ray.rllib.core.models.catalog import Catalog
from ray.rllib.core.models.base import Encoder, ActorCriticEncoder
from ray.rllib.core.models.configs import MLPHeadConfig, MLPEncoderConfig, ActorCriticEncoderConfig
from ray.rllib.algorithms.ppo.torch.default_ppo_torch_rl_module import DefaultPPOTorchRLModule
from ray.rllib.algorithms.ppo.ppo_catalog import PPOCatalog
from ray.rllib.utils.numpy import convert_to_numpy
import gymnasium as gym
import numpy as np
import torch
import torch.nn as nn

class DictAPPOCatalog(PPOCatalog):
    """自定义 Catalog 处理字典观测空间"""
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def _get_encoder_config(
            self,) -> MLPEncoderConfig:
        """为字典观测空间创建多分支编码器配置"""
        # 跳过父类方法，不调用 super()
        if not isinstance(self.observation_space, gym.spaces.Dict):
            return super()._get_encoder_config(framework)
        
        # 创建多分支配置
        branch_configs = {}
        for key, subspace in self.observation_space.spaces.items():
            branch_configs[key] = MLPEncoderConfig(
                input_dim=subspace.shape[0],
                hidden_layer_dims=self.model_config_dict["fcnet_hiddens"],
                hidden_layer_activation=self.model_config_dict["fcnet_activation"],
                output_dim=self.model_config_dict["fcnet_hiddens"][-1],
                output_activation=self.model_config_dict["post_fcnet_activation"],
            )
        
        # 返回为字典空间定制的编码器配置
        return DictEncoderConfig(branch_configs=branch_configs)
    
    def build_actor_critic_encoder(self, framework: str) -> ActorCriticEncoder:
        """构建处理字典空间的演员-评论家编码器"""
        if not isinstance(self.observation_space, gym.spaces.Dict):
            return super().build_actor_critic_encoder(framework)
        
        encoder_config = self._get_encoder_config(framework)
        encoder = encoder_config.build(framework)
        
        return DictActorCriticEncoder(
            encoder=encoder,
            action_space=self.action_space,
            model_config_dict=self.model_config_dict,
        )

class DictEncoderConfig(MLPEncoderConfig):
    """字典观测的编码器配置"""
    
    def __init__(self, branch_configs: dict):
        super().__init__(input_dim=None)
        self.branch_configs = branch_configs
        
    def build(self, framework: str = "torch") -> "DictEncoder":
        return DictEncoder(self)

class DictEncoder(Encoder):
    """处理字典观测的编码器模块"""
    
    def __init__(self, config: DictEncoderConfig):
        super().__init__()
        self.branches = nn.ModuleDict()
        
        # 为观测字典的每个键创建独立的MLP分支
        for key, branch_config in config.branch_configs.items():
            mlp_config = MLPEncoderConfig(
                input_dim=branch_config.input_dim,
                hidden_layer_dims=branch_config.hidden_layer_dims,
                hidden_layer_activation=branch_config.hidden_layer_activation,
                output_dim=branch_config.output_dim,
                output_activation=branch_config.output_activation
            )
            self.branches[key] = mlp_config.build(framework="torch")
        
        # 计算总的输出维度
        self.output_dim = sum(
            branch_config.output_dim 
            for branch_config in config.branch_configs.values()
        )
    
    @property
    def output_dim(self) -> int:
        return self._output_dim
    
    @output_dim.setter
    def output_dim(self, value):
        self._output_dim = value
        
    def _forward(self, inputs: dict, **kwargs) -> torch.Tensor:
        branch_outputs = []
        
        # 处理每个观测分支
        for key, branch in self.branches.items():
            # 确保输入中包含所有键
            if key not in inputs:
                raise ValueError(f"Observation missing required key: {key}")
                
            # 获取输入并确保正确的形状
            obs_data = inputs[key]
            if not isinstance(obs_data, torch.Tensor):
                obs_data = torch.as_tensor(obs_data, dtype=torch.float32)
                
            # 通过分支处理
            branch_out = branch(obs_data)
            branch_outputs.append(branch_out)
        
        # 沿着特征维度拼接所有分支输出
        return torch.cat(branch_outputs, dim=-1)

class DictActorCriticEncoder(ActorCriticEncoder):
    """处理字典观测空间的演员-评论家编码器"""
    
    def __init__(self, encoder: Encoder, *args, **kwargs):
        super().__init__(encoder, *args, **kwargs)
    
    def __call__(self, inputs: dict, **kwargs) -> dict:
        # 通过共享编码器处理字典观测
        encoder_out = self.encoder(inputs)
        
        # 返回统一格式的编码输出
        return {
            ENCODER_OUT: encoder_out,
            "actor_out": encoder_out,
            "critic_out": encoder_out,
        }

class DictAPPOTorchRLModule(DefaultPPOTorchRLModule):
    """处理字典观测的自定义PPO RL模块"""
    
    def setup(self):
        # 使用我们的自定义Catalog
        self.config.catalog_class = DictAPPOCatalog
        self.config.catalog = self.config.get_catalog()
        catalog = self.config.catalog
        
        # 确保共享编码器用于演员和评论家
        encoder_config = ActorCriticEncoderConfig()

        # 构建共享编码器
        self.encoder = catalog.build_actor_critic_encoder(
            encoder_config=encoder_config,
            framework=self.framework
        )

        # 构建演员和评论家头部
        actor_head = catalog.build_actor_head(framework=self.framework)
        critic_head = catalog.build_critic_head(framework=self.framework)

        # 配置值函数共享
        self.pi = actor_head(self.encoder.output_dim)
        self.vf = critic_head(self.encoder.output_dim)


from .device import Device
from .keyboard import Keyboard

try:
    from .spacemouse import SpaceMouse
except ImportError:
    print(
        """Unable to load module hid, required to interface with SpaceMouse.\n
           Only macOS is officially supported. Install the additional\n
           requirements with `pip install -r requirements-extra.txt`"""
    )



import numpy as np
from gymnasium.core import ObsType
from typing import Optional, Any, SupportsFloat
from gymnasium import spaces
import datetime


class OrcaGymAsyncAgent:
    def __init__(self, 
                 env_id: str,
                 agent_name: str, 
                 task: str,
                 max_episode_steps: int,
                 dt: float,
                 **kwargs):
        
        self._env_id = env_id
        self._agent_name = agent_name
        self._task = task
        self._max_episode_steps = max_episode_steps
        self._current_episode_step = 0
        self._dt = dt

        self._joint_names = None
        self._actuator_names = None
        self._site_names = None
        self._sensor_names = None

        self._actuator_type = None
        self._action_scale = None
        self._action_scale_mask = None
        self._ctrl_range = None
        self._ctrl_delta_range = None
        self._ctrl_range_low = None
        self._ctrl_range_high = None
        self._ctrl_start = None
        self._neutral_joint_values = None

        self._ctrl = None
        self._nu = None
        self._nq = None
        self._nv = None

    @property
    def dt(self) -> float:
        return self._dt

    @property
    def name(self) -> str:
        return self._agent_name

    def name_space(self, name : str) -> str:
        return f"{self._agent_name}_{name}"
    
    def name_space_list(self, names : list[str]) -> list[str]:
        return [self.name_space(name) for name in names]
    
    @property
    def joint_names(self) -> list[str]:
        return self._joint_names
    
    @property
    def actuator_names(self) -> list[str]:
        return self._actuator_names
    
    @property
    def site_names(self) -> list[str]:
        return self._site_names
    
    @property
    def sensor_names(self) -> list[str]:
        return self._sensor_names

    @property
    def nu(self) -> int:
        return self._nu
    
    @property
    def nq(self) -> int:
        return self._nq
    
    @property
    def nv(self) -> int:
        return self._nv

    @property
    def truncated(self) -> bool:
        return self._current_episode_step >= self._max_episode_steps

    @property
    def ctrl_start(self) -> int:
        return self._ctrl_start

    @property
    def action_range(self) -> np.ndarray:
        raise NotImplementedError

    @property
    def kps(self) -> float:
        raise NotImplementedError
    
    @property
    def kds(self) -> float:
        raise NotImplementedError

    def get_obs(self, **kwargs):
        raise NotImplementedError

    def init_ctrl_info(self, actuator_dict, joint_dict) -> None:
        ctrl_range_list = []
        ctrl_delta_range_list = []
        torques_range_list = []
        for i, joint_name in enumerate(self._joint_names):
            if i == 0:
                continue    # pass the first joint, which is the free joint
            # matain the order of actuators
            ctrl_range_list.append(np.array(joint_dict[joint_name]['Range']).flatten())
            ctrl_range_width = ctrl_range_list[-1][1] - ctrl_range_list[-1][0]
            ctrl_delta_range_list.append([-ctrl_range_width/2, ctrl_range_width/2])

        for actuator_name in self._actuator_names:
            torques_range_list.append(np.array(actuator_dict[actuator_name]['CtrlRange']).flatten())


        self._ctrl_start = actuator_dict[self._actuator_names[0]]['ActuatorId']

        self._torques_range = np.array(torques_range_list)
        self._torques_limit = np.array([abs(self._torques_range[i][1]) * self._soft_torque_limit for i in range(len(self._torques_range))])

        self._ctrl_range = np.array(ctrl_range_list)
        self._joint_qpos_limit = np.array([[self._ctrl_range[i][0] * self._soft_joint_qpos_limit, 
                                           self._ctrl_range[i][1] * self._soft_joint_qpos_limit] 
                                           for i in range(len(self._ctrl_range))])

        self._joint_qvel_limit = np.array([self._joint_qvel_range[i] * self._soft_joint_qvel_limit for i in range(len(self._joint_qvel_range))])

        # print("ctrl_range: ", self._ctrl_range)
        # print("soft_joint_qpos_limit: ", self._soft_joint_qpos_limit)
        # print("joint_qpos_limit: ", self._joint_qpos_limit)

        self._ctrl_delta_range = np.array(ctrl_delta_range_list)


        self._ctrl_range_low = np.array([range[0] for range in self._ctrl_range])
        self._ctrl_range_high = np.array([range[1] for range in self._ctrl_range])

    def get_ctrl_info(self) -> dict:
        return {
            "actuator_type": self._actuator_type,
            "action_scale": self._action_scale,
            "action_scale_mask": self._action_scale_mask,
            "ctrl_range": self._ctrl_range,
            "ctrl_delta_range": self._ctrl_delta_range,
            "ctrl_range_low": self._ctrl_range_low,
            "ctrl_range_high": self._ctrl_range_high,
            "ctrl_start": self._ctrl_start,
            "ctrl_end": self._ctrl_start + len(self._actuator_names),
            "neutral_joint_values": self._neutral_joint_values,
        }
    
    def init_joint_index(self, qpos_offset, qvel_offset, qacc_offset, qpos_length, qvel_length, qacc_length):
        """
        Joint index is specific to the agent and is defined in the subclass.
        """
        raise NotImplementedError
    
    def set_action_space(self) -> None:
        """
        Action space is specific to the agent and is defined in the subclass.
        """
        raise NotImplementedError
    
    def on_step(self, action, **kwargs):
        """
        Called after each step in the environment.
        Implement this method in the subclass to perform additional operations.
        """
        raise NotImplementedError

    def step(self, action, **kwargs):
        self._current_episode_step += 1
        step_info = self.on_step(action, **kwargs)
        return self._ctrl, step_info

    def on_reset(self, **kwargs):
        """
        Called after each reset in the environment.
        Implement this method in the subclass to perform additional operations.
        """
        raise NotImplementedError

    def reset(self, np_random : np.random.Generator, **kwargs):
        self._current_episode_step = 0
        self._np_random = np_random
        reset_info = self.on_reset(**kwargs)
        return reset_info

    def is_success(self, achieved_goal, desired_goal) -> np.float32:
        raise NotImplementedError

    def is_terminated(self, achieved_goal, desired_goal) -> bool:
        raise NotImplementedError
        
    def compute_reward(self, achieved_goal, desired_goal) -> SupportsFloat:
        raise NotImplementedError
        
    def set_init_state(self, joint_qpos: dict, init_site_pos_quat: dict) -> None:
        raise NotImplementedError
    
    def get_action_size(self) -> int:
        """
        Action size can be overridden in the subclass.
        In most of cases, this is the number of actuators in the robot.
        But in some cases, the action size may be different.
        """        
        return self._nu

    def compute_torques(self, qpos_buffer : np.ndarray, qvel_buffer : np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def setup_curriculum(self, curriculum : str) -> None:
        raise NotImplementedError


"""
This file contains utility classes and functions for logging to stdout and stderr
Adapted from robomimic: https://github.com/ARISE-Initiative/robomimic/blob/master/robomimic/utils/log_utils.py
"""
import logging
import os
import time

from termcolor import colored

import orca_gym.adapters.robosuite.macros as macros

LEVEL_COLORS = {
    logging.DEBUG: "green",
    logging.INFO: "green",
    logging.WARNING: "yellow",
    logging.ERROR: "red",
    logging.CRITICAL: "red",
}

FORMAT_STR = {"file": "[robosuite %(levelname)s - %(asctime)s] ", "console": "[robosuite %(levelname)s] "}

MESSAGE_STR = "%(message)s (%(filename)s:%(lineno)d)"


class FileFormatter(logging.Formatter):
    """Formatter class of logging for file logging."""

    FORMATS = {
        levelno: colored(FORMAT_STR["file"], color, attrs=["bold"]) + MESSAGE_STR
        for (levelno, color) in LEVEL_COLORS.items()
    }

    def format(self, record):
        """Apply custom fomatting on LogRecord object record."""
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, "%Y-%m-%d %H:%M:%S")
        return formatter.format(record)


class ConsoleFormatter(logging.Formatter):
    """Formatter class of logging for console logging."""

    FORMATS = {
        logging.DEBUG: FORMAT_STR["console"] + MESSAGE_STR,
        logging.INFO: "%(message)s",
        logging.WARNING: colored(FORMAT_STR["console"], "yellow", attrs=["bold"]) + MESSAGE_STR,
        logging.ERROR: colored(FORMAT_STR["console"], "red", attrs=["bold"]) + MESSAGE_STR,
        logging.CRITICAL: colored(FORMAT_STR["console"], "red", attrs=["bold", "reverse"]) + MESSAGE_STR,
    }

    def format(self, record):
        """Apply custom fomatting on LogRecord object record."""
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class DefaultLogger:
    """Default logger class in orca_gym.adapters.robosuite.codebase."""

    def __init__(self, logger_name="robosuite_logs", console_logging_level="INFO", file_logging_level=None):
        """
        Args:
            logger_name (str, optional): logger name. Defaults to "robosuite_logs".
            console_logging_level (str, optional): logging level for console logging. Defaults to "INFO".
            file_logging_level (_type_, optional): logging level for file logging. Defaults to None.
        """
        self.logger_name = logger_name
        logger = logging.getLogger(self.logger_name)

        if file_logging_level is not None:
            time_str = str(time.time()).replace(".", "_")
            log_file_path = "/tmp/robosuite_{}_{}.log".format(time_str, os.getpid())
            fh = logging.FileHandler(log_file_path)
            print(colored("[robosuite]: Saving logs to {}".format(log_file_path), "yellow"))
            fh.setLevel(logging.getLevelName(file_logging_level))
            file_formatter = FileFormatter()
            fh.setFormatter(file_formatter)
            logger.addHandler(fh)

        if console_logging_level is not None:
            ch = logging.StreamHandler()
            ch.setLevel(logging.getLevelName(console_logging_level))
            console_formatter = ConsoleFormatter()
            ch.setFormatter(console_formatter)
            logger.addHandler(ch)

    def get_logger(self):
        """_summary_

        Returns:
            DefaultLogger: The retrieved logger whose name equals self.logger_name
        """
        logger = logging.getLogger(self.logger_name)
        return logger


ROBOSUITE_DEFAULT_LOGGER = DefaultLogger(
    console_logging_level=macros.CONSOLE_LOGGING_LEVEL,
    file_logging_level=macros.FILE_LOGGING_LEVEL,
).get_logger()



from .core.orca_gym import OrcaGymBase
from .core.orca_gym_model import OrcaGymModel
from .core.orca_gym_data import OrcaGymData
from .core.orca_gym_opt_config import OrcaGymOptConfig
from .core.orca_gym_local import OrcaGymLocal
from .core.orca_gym_remote import OrcaGymRemote
import orca_gym.protos.mjc_message_pb2 as mjc_message_pb2
import orca_gym.protos.mjc_message_pb2_grpc as mjc_message_pb2_grpc

__all__ = ['OrcaGymBase', 'OrcaGymModel', 'OrcaGymData', 'OrcaGymOptConfig', 'OrcaGymLocal', 'OrcaGymRemote', 'mjc_message_pb2', 'mjc_message_pb2_grpc']


"""
Modder classes used for domain randomization. Largely based off of the mujoco-py
implementation below.

https://github.com/openai/mujoco-py/blob/1fe312b09ae7365f0dd9d4d0e453f8da59fae0bf/mujoco_py/modder.py
"""

import copy
import os
from collections import defaultdict

import numpy as np
from PIL import Image

import robosuite
import orca_gym.adapters.robosuite.utils.transform_utils as trans
from orca_gym.adapters.robosuite.utils.binding_utils import MjRenderContextOffscreen


class BaseModder:
    """
    Base class meant to modify simulation attributes mid-sim.

    Using @random_state ensures that sampling here won't be affected
    by sampling that happens outside of the modders.

    Args:
        sim (MjSim): simulation object

        random_state (RandomState): instance of np.random.RandomState, specific
            seed used to randomize these modifications without impacting other
            numpy seeds / randomizations
    """

    def __init__(self, sim, random_state=None):
        self.sim = sim
        if random_state is None:
            # default to global RandomState instance
            self.random_state = np.random.mtrand._rand
        else:
            self.random_state = random_state

    def update_sim(self, sim):
        """
        Setter function to update internal sim variable

        Args:
            sim (MjSim): MjSim object
        """
        self.sim = sim

    @property
    def model(self):
        """
        Returns:
            MjModel: Mujoco sim model
        """
        # Available for quick convenience access
        return self.sim.model


class LightingModder(BaseModder):
    """
    Modder to modify lighting within a Mujoco simulation.

    Args:
        sim (MjSim): MjSim object

        random_state (RandomState): instance of np.random.RandomState

        light_names (None or list of str): list of lights to use for randomization. If not provided, all
            lights in the model are randomized.

        randomize_position (bool): If True, randomizes position of lighting

        randomize_direction (bool): If True, randomizes direction of lighting

        randomize_specular (bool): If True, randomizes specular attribute of lighting

        randomize_ambient (bool): If True, randomizes ambient attribute of lighting

        randomize_diffuse (bool): If True, randomizes diffuse attribute of lighting

        randomize_active (bool): If True, randomizes active nature of lighting

        position_perturbation_size (float): Magnitude of position randomization

        direction_perturbation_size (float): Magnitude of direction randomization

        specular_perturbation_size (float): Magnitude of specular attribute randomization

        ambient_perturbation_size (float): Magnitude of ambient attribute randomization

        diffuse_perturbation_size (float): Magnitude of diffuse attribute randomization
    """

    def __init__(
        self,
        sim,
        random_state=None,
        light_names=None,
        randomize_position=True,
        randomize_direction=True,
        randomize_specular=True,
        randomize_ambient=True,
        randomize_diffuse=True,
        randomize_active=True,
        position_perturbation_size=0.1,
        direction_perturbation_size=0.35,  # 20 degrees
        specular_perturbation_size=0.1,
        ambient_perturbation_size=0.1,
        diffuse_perturbation_size=0.1,
    ):
        super().__init__(sim, random_state=random_state)

        if light_names is None:
            light_names = self.sim.model.light_names
        self.light_names = light_names

        self.randomize_position = randomize_position
        self.randomize_direction = randomize_direction
        self.randomize_specular = randomize_specular
        self.randomize_ambient = randomize_ambient
        self.randomize_diffuse = randomize_diffuse
        self.randomize_active = randomize_active

        self.position_perturbation_size = position_perturbation_size
        self.direction_perturbation_size = direction_perturbation_size
        self.specular_perturbation_size = specular_perturbation_size
        self.ambient_perturbation_size = ambient_perturbation_size
        self.diffuse_perturbation_size = diffuse_perturbation_size

        self.save_defaults()

    def save_defaults(self):
        """
        Uses the current MjSim state and model to save default parameter values.
        """
        self._defaults = {k: {} for k in self.light_names}
        for name in self.light_names:
            self._defaults[name]["pos"] = np.array(self.get_pos(name))
            self._defaults[name]["dir"] = np.array(self.get_dir(name))
            self._defaults[name]["specular"] = np.array(self.get_specular(name))
            self._defaults[name]["ambient"] = np.array(self.get_ambient(name))
            self._defaults[name]["diffuse"] = np.array(self.get_diffuse(name))
            self._defaults[name]["active"] = self.get_active(name)

    def restore_defaults(self):
        """
        Reloads the saved parameter values.
        """
        for name in self.light_names:
            self.set_pos(name, self._defaults[name]["pos"])
            self.set_dir(name, self._defaults[name]["dir"])
            self.set_specular(name, self._defaults[name]["specular"])
            self.set_ambient(name, self._defaults[name]["ambient"])
            self.set_diffuse(name, self._defaults[name]["diffuse"])
            self.set_active(name, self._defaults[name]["active"])

    def randomize(self):
        """
        Randomizes all requested lighting values within the sim
        """
        for name in self.light_names:
            if self.randomize_position:
                self._randomize_position(name)

            if self.randomize_direction:
                self._randomize_direction(name)

            if self.randomize_specular:
                self._randomize_specular(name)

            if self.randomize_ambient:
                self._randomize_ambient(name)

            if self.randomize_diffuse:
                self._randomize_diffuse(name)

            if self.randomize_active:
                self._randomize_active(name)

    def _randomize_position(self, name):
        """
        Helper function to randomize position of a specific light source

        Args:
            name (str): Name of the lighting source to randomize for
        """
        delta_pos = self.random_state.uniform(
            low=-self.position_perturbation_size,
            high=self.position_perturbation_size,
            size=3,
        )
        self.set_pos(
            name,
            self._defaults[name]["pos"] + delta_pos,
        )

    def _randomize_direction(self, name):
        """
        Helper function to randomize direction of a specific light source

        Args:
            name (str): Name of the lighting source to randomize for
        """
        # sample a small, random axis-angle delta rotation
        random_axis, random_angle = trans.random_axis_angle(
            angle_limit=self.direction_perturbation_size, random_state=self.random_state
        )
        random_delta_rot = trans.quat2mat(trans.axisangle2quat(random_axis * random_angle))

        # rotate direction by this delta rotation and set the new direction
        new_dir = random_delta_rot.dot(self._defaults[name]["dir"])
        self.set_dir(
            name,
            new_dir,
        )

    def _randomize_specular(self, name):
        """
        Helper function to randomize specular attribute of a specific light source

        Args:
            name (str): Name of the lighting source to randomize for
        """
        delta = self.random_state.uniform(
            low=-self.specular_perturbation_size,
            high=self.specular_perturbation_size,
            size=3,
        )
        self.set_specular(
            name,
            self._defaults[name]["specular"] + delta,
        )

    def _randomize_ambient(self, name):
        """
        Helper function to randomize ambient attribute of a specific light source

        Args:
            name (str): Name of the lighting source to randomize for
        """
        delta = self.random_state.uniform(
            low=-self.ambient_perturbation_size,
            high=self.ambient_perturbation_size,
            size=3,
        )
        self.set_ambient(
            name,
            self._defaults[name]["ambient"] + delta,
        )

    def _randomize_diffuse(self, name):
        """
        Helper function to randomize diffuse attribute of a specific light source

        Args:
            name (str): Name of the lighting source to randomize for
        """
        delta = self.random_state.uniform(
            low=-self.diffuse_perturbation_size,
            high=self.diffuse_perturbation_size,
            size=3,
        )
        self.set_diffuse(
            name,
            self._defaults[name]["diffuse"] + delta,
        )

    def _randomize_active(self, name):
        """
        Helper function to randomize active nature of a specific light source

        Args:
            name (str): Name of the lighting source to randomize for
        """
        active = int(self.random_state.uniform() > 0.5)
        self.set_active(name, active)

    def get_pos(self, name):
        """
        Grabs position of a specific light source

        Args:
            name (str): Name of the lighting source

        Returns:
            np.array: (x,y,z) position of lighting source

        Raises:
            AssertionError: Invalid light name
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        return self.model.light_pos[lightid]

    def set_pos(self, name, value):
        """
        Sets position of a specific light source

        Args:
            name (str): Name of the lighting source
            value (np.array): (x,y,z) position to set lighting source to

        Raises:
            AssertionError: Invalid light name
            AssertionError: Invalid @value
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        value = list(value)
        assert len(value) == 3, "Expected 3-dim value, got %s" % value

        self.model.light_pos[lightid] = value

    def get_dir(self, name):
        """
        Grabs direction of a specific light source

        Args:
            name (str): Name of the lighting source

        Returns:
            np.array: (x,y,z) direction of lighting source

        Raises:
            AssertionError: Invalid light name
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        return self.model.light_dir[lightid]

    def set_dir(self, name, value):
        """
        Sets direction of a specific light source

        Args:
            name (str): Name of the lighting source
            value (np.array): (ax,ay,az) direction to set lighting source to

        Raises:
            AssertionError: Invalid light name
            AssertionError: Invalid @value
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        value = list(value)
        assert len(value) == 3, "Expected 3-dim value, got %s" % value

        self.model.light_dir[lightid] = value

    def get_active(self, name):
        """
        Grabs active nature of a specific light source

        Args:
            name (str): Name of the lighting source

        Returns:
            int: Whether light source is active (1) or not (0)

        Raises:
            AssertionError: Invalid light name
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        return self.model.light_active[lightid]

    def set_active(self, name, value):
        """
        Sets active nature of a specific light source

        Args:
            name (str): Name of the lighting source
            value (int): Whether light source is active (1) or not (0)

        Raises:
            AssertionError: Invalid light name
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        self.model.light_active[lightid] = value

    def get_specular(self, name):
        """
        Grabs specular attribute of a specific light source

        Args:
            name (str): Name of the lighting source

        Returns:
            np.array: (r,g,b) specular color of lighting source

        Raises:
            AssertionError: Invalid light name
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        return self.model.light_specular[lightid]

    def set_specular(self, name, value):
        """
        Sets specular attribute of a specific light source

        Args:
            name (str): Name of the lighting source
            value (np.array): (r,g,b) specular color to set lighting source to

        Raises:
            AssertionError: Invalid light name
            AssertionError: Invalid @value
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        value = list(value)
        assert len(value) == 3, "Expected 3-dim value, got %s" % value

        self.model.light_specular[lightid] = value

    def get_ambient(self, name):
        """
        Grabs ambient attribute of a specific light source

        Args:
            name (str): Name of the lighting source

        Returns:
            np.array: (r,g,b) ambient color of lighting source

        Raises:
            AssertionError: Invalid light name
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        return self.model.light_ambient[lightid]

    def set_ambient(self, name, value):
        """
        Sets ambient attribute of a specific light source

        Args:
            name (str): Name of the lighting source
            value (np.array): (r,g,b) ambient color to set lighting source to

        Raises:
            AssertionError: Invalid light name
            AssertionError: Invalid @value
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        value = list(value)
        assert len(value) == 3, "Expected 3-dim value, got %s" % value

        self.model.light_ambient[lightid] = value

    def get_diffuse(self, name):
        """
        Grabs diffuse attribute of a specific light source

        Args:
            name (str): Name of the lighting source

        Returns:
            np.array: (r,g,b) diffuse color of lighting source

        Raises:
            AssertionError: Invalid light name
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        return self.model.light_diffuse[lightid]

    def set_diffuse(self, name, value):
        """
        Sets diffuse attribute of a specific light source

        Args:
            name (str): Name of the lighting source
            value (np.array): (r,g,b) diffuse color to set lighting source to

        Raises:
            AssertionError: Invalid light name
            AssertionError: Invalid @value
        """
        lightid = self.get_lightid(name)
        assert lightid > -1, "Unkwnown light %s" % name

        value = list(value)
        assert len(value) == 3, "Expected 3-dim value, got %s" % value

        self.model.light_diffuse[lightid] = value

    def get_lightid(self, name):
        """
        Grabs unique id number of a specific light source

        Args:
            name (str): Name of the lighting source

        Returns:
            int: id of lighting source. -1 if not found
        """
        return self.model.light_name2id(name)


class CameraModder(BaseModder):
    """
    Modder for modifying camera attributes in mujoco sim

    Args:
        sim (MjSim): MjSim object

        random_state (None or RandomState): instance of np.random.RandomState

        camera_names (None or list of str): list of camera names to use for randomization. If not provided,
            all cameras are used for randomization.

        randomize_position (bool): if True, randomize camera position

        randomize_rotation (bool): if True, randomize camera rotation

        randomize_fovy (bool): if True, randomize camera fovy

        position_perturbation_size (float): size of camera position perturbations to each dimension

        rotation_perturbation_size (float): magnitude of camera rotation perturbations in axis-angle.
            Default corresponds to around 5 degrees.

        fovy_perturbation_size (float): magnitude of camera fovy perturbations (corresponds to focusing)

    Raises:
        AssertionError: [No randomization selected]
    """

    def __init__(
        self,
        sim,
        random_state=None,
        camera_names=None,
        randomize_position=True,
        randomize_rotation=True,
        randomize_fovy=True,
        position_perturbation_size=0.01,
        rotation_perturbation_size=0.087,
        fovy_perturbation_size=5.0,
    ):
        super().__init__(sim, random_state=random_state)

        assert randomize_position or randomize_rotation or randomize_fovy

        if camera_names is None:
            camera_names = self.sim.model.camera_names
        self.camera_names = camera_names

        self.randomize_position = randomize_position
        self.randomize_rotation = randomize_rotation
        self.randomize_fovy = randomize_fovy

        self.position_perturbation_size = position_perturbation_size
        self.rotation_perturbation_size = rotation_perturbation_size
        self.fovy_perturbation_size = fovy_perturbation_size

        self.save_defaults()

    def save_defaults(self):
        """
        Uses the current MjSim state and model to save default parameter values.
        """
        self._defaults = {k: {} for k in self.camera_names}
        for camera_name in self.camera_names:
            self._defaults[camera_name]["pos"] = np.array(self.get_pos(camera_name))
            self._defaults[camera_name]["quat"] = np.array(self.get_quat(camera_name))
            self._defaults[camera_name]["fovy"] = self.get_fovy(camera_name)

    def restore_defaults(self):
        """
        Reloads the saved parameter values.
        """
        for camera_name in self.camera_names:
            self.set_pos(camera_name, self._defaults[camera_name]["pos"])
            self.set_quat(camera_name, self._defaults[camera_name]["quat"])
            self.set_fovy(camera_name, self._defaults[camera_name]["fovy"])

    def randomize(self):
        """
        Randomizes all requested camera values within the sim
        """
        for camera_name in self.camera_names:
            if self.randomize_position:
                self._randomize_position(camera_name)

            if self.randomize_rotation:
                self._randomize_rotation(camera_name)

            if self.randomize_fovy:
                self._randomize_fovy(camera_name)

    def _randomize_position(self, name):
        """
        Helper function to randomize position of a specific camera

        Args:
            name (str): Name of the camera to randomize for
        """
        delta_pos = self.random_state.uniform(
            low=-self.position_perturbation_size,
            high=self.position_perturbation_size,
            size=3,
        )
        self.set_pos(
            name,
            self._defaults[name]["pos"] + delta_pos,
        )

    def _randomize_rotation(self, name):
        """
        Helper function to randomize orientation of a specific camera

        Args:
            name (str): Name of the camera to randomize for
        """
        # sample a small, random axis-angle delta rotation
        random_axis, random_angle = trans.random_axis_angle(
            angle_limit=self.rotation_perturbation_size, random_state=self.random_state
        )
        random_delta_rot = trans.quat2mat(trans.axisangle2quat(random_axis * random_angle))

        # compute new rotation and set it
        base_rot = trans.quat2mat(trans.convert_quat(self._defaults[name]["quat"], to="xyzw"))
        new_rot = random_delta_rot.T.dot(base_rot)
        new_quat = trans.convert_quat(trans.mat2quat(new_rot), to="wxyz")
        self.set_quat(
            name,
            new_quat,
        )

    def _randomize_fovy(self, name):
        """
        Helper function to randomize fovy of a specific camera

        Args:
            name (str): Name of the camera to randomize for
        """
        delta_fovy = self.random_state.uniform(
            low=-self.fovy_perturbation_size,
            high=self.fovy_perturbation_size,
        )
        self.set_fovy(
            name,
            self._defaults[name]["fovy"] + delta_fovy,
        )

    def get_fovy(self, name):
        """
        Grabs fovy of a specific camera

        Args:
            name (str): Name of the camera

        Returns:
            float: vertical field of view of the camera, expressed in degrees

        Raises:
            AssertionError: Invalid camera name
        """
        camid = self.get_camid(name)
        assert camid > -1, "Unknown camera %s" % name
        return self.model.cam_fovy[camid]

    def set_fovy(self, name, value):
        """
        Sets fovy of a specific camera

        Args:
            name (str): Name of the camera
            value (float): vertical field of view of the camera, expressed in degrees

        Raises:
            AssertionError: Invalid camera name
            AssertionError: Invalid value
        """
        camid = self.get_camid(name)
        assert 0 < value < 180
        assert camid > -1, "Unknown camera %s" % name
        self.model.cam_fovy[camid] = value

    def get_quat(self, name):
        """
        Grabs orientation of a specific camera

        Args:
            name (str): Name of the camera

        Returns:
            np.array: (w,x,y,z) orientation of the camera, expressed in quaternions

        Raises:
            AssertionError: Invalid camera name
        """
        camid = self.get_camid(name)
        assert camid > -1, "Unknown camera %s" % name
        return self.model.cam_quat[camid]

    def set_quat(self, name, value):
        """
        Sets orientation of a specific camera

        Args:
            name (str): Name of the camera
            value (np.array): (w,x,y,z) orientation of the camera, expressed in quaternions

        Raises:
            AssertionError: Invalid camera name
            AssertionError: Invalid value
        """
        value = list(value)
        assert len(value) == 4, "Expectd value of length 4, instead got %s" % value
        camid = self.get_camid(name)
        assert camid > -1, "Unknown camera %s" % name
        self.model.cam_quat[camid] = value

    def get_pos(self, name):
        """
        Grabs position of a specific camera

        Args:
            name (str): Name of the camera

        Returns:
            np.array: (x,y,z) position of the camera

        Raises:
            AssertionError: Invalid camera name
        """
        camid = self.get_camid(name)
        assert camid > -1, "Unknown camera %s" % name
        return self.model.cam_pos[camid]

    def set_pos(self, name, value):
        """
        Sets position of a specific camera

        Args:
            name (str): Name of the camera
            value (np.array): (x,y,z) position of the camera

        Raises:
            AssertionError: Invalid camera name
            AssertionError: Invalid value
        """
        value = list(value)
        assert len(value) == 3, "Expected value of length 3, instead got %s" % value
        camid = self.get_camid(name)
        assert camid > -1
        self.model.cam_pos[camid] = value

    def get_camid(self, name):
        """
        Grabs unique id number of a specific camera

        Args:
            name (str): Name of the camera

        Returns:
            int: id of camera. -1 if not found
        """
        return self.model.camera_name2id(name)


class TextureModder(BaseModder):
    """
    Modify textures in model. Example use:
        sim = MjSim(...)
        modder = TextureModder(sim)
        modder.whiten_materials()  # ensures materials won't impact colors
        modder.set_checker('some_geom', (255, 0, 0), (0, 0, 0))
        modder.rand_all('another_geom')

    Note: in order for the textures to take full effect, you'll need to set
    the rgba values for all materials to [1, 1, 1, 1], otherwise the texture
    colors will be modulated by the material colors. Call the
    `whiten_materials` helper method to set all material colors to white.

    Args:
        sim (MjSim): MjSim object

        random_state (RandomState): instance of np.random.RandomState

        geom_names ([string]): list of geom names to use for randomization. If not provided,
            all geoms are used for randomization.

        randomize_local (bool): if True, constrain RGB color variations to be close to the
            original RGB colors per geom and texture. Otherwise, RGB color values will
            be sampled uniformly at random.

        randomize_material (bool): if True, randomizes material properties associated with a
            given texture (reflectance, shininess, specular)

        local_rgb_interpolation (float): determines the size of color variations from
            the base geom colors when @randomize_local is True.

        local_material_interpolation (float): determines the size of material variations from
            the base material when @randomize_local and @randomize_material are both True.

        texture_variations (list of str): a list of texture variation strings. Each string
            must be either 'rgb', 'checker', 'noise', or 'gradient' and corresponds to
            a specific kind of texture randomization. For each geom that has a material
            and texture, a random variation from this list is sampled and applied.

        randomize_skybox (bool): if True, apply texture variations to the skybox as well.
    """

    def __init__(
        self,
        sim,
        random_state=None,
        geom_names=None,
        randomize_local=False,
        randomize_material=False,
        local_rgb_interpolation=0.1,
        local_material_interpolation=0.2,
        texture_variations=("rgb", "checker", "noise", "gradient"),
        randomize_skybox=True,
    ):
        super().__init__(sim, random_state=random_state)

        if geom_names is None:
            geom_names = self.sim.model.geom_names
        self.geom_names = geom_names

        self.randomize_local = randomize_local
        self.randomize_material = randomize_material
        self.local_rgb_interpolation = local_rgb_interpolation
        self.local_material_interpolation = local_material_interpolation
        self.texture_variations = list(texture_variations)
        self.randomize_skybox = randomize_skybox

        self._all_texture_variation_callbacks = {
            "rgb": self.rand_rgb,
            "checker": self.rand_checker,
            "noise": self.rand_noise,
            "gradient": self.rand_gradient,
        }
        self._texture_variation_callbacks = {
            k: self._all_texture_variation_callbacks[k] for k in self.texture_variations
        }

        self.save_defaults()

    def save_defaults(self):
        """
        Uses the current MjSim state and model to save default parameter values.
        """
        self.textures = [Texture(self.model, i) for i in range(self.model.ntex)]
        # self._build_tex_geom_map()

        # save copy of original texture bitmaps
        self._default_texture_bitmaps = [np.array(text.bitmap) for text in self.textures]

        # These matrices will be used to rapidly synthesize
        # checker pattern bitmaps
        self._cache_checker_matrices()

        self._defaults = {k: {} for k in self.geom_names}
        if self.randomize_skybox:
            self._defaults["skybox"] = {}
        for name in self.geom_names:
            if self._check_geom_for_texture(name):
                # store the texture bitmap for this geom
                tex_id = self._name_to_tex_id(name)
                self._defaults[name]["texture"] = self._default_texture_bitmaps[tex_id]
                # store material properties as well (in tuple (reflectance, shininess, specular) form)
                self._defaults[name]["material"] = self.get_material(name)
            else:
                # store geom color
                self._defaults[name]["rgb"] = np.array(self.get_geom_rgb(name))

        if self.randomize_skybox:
            tex_id = self._name_to_tex_id("skybox")
            self._defaults["skybox"]["texture"] = self._default_texture_bitmaps[tex_id]

    def restore_defaults(self):
        """
        Reloads the saved parameter values.
        """
        for name in self.geom_names:
            if self._check_geom_for_texture(name):
                self.set_texture(name, self._defaults[name]["texture"], perturb=False)
                self.set_material(name, self._defaults[name]["material"], perturb=False)
            else:
                self.set_geom_rgb(name, self._defaults[name]["rgb"])

        if self.randomize_skybox:
            self.set_texture("skybox", self._defaults["skybox"]["texture"], perturb=False)

    def randomize(self):
        """
        Overrides mujoco-py implementation to also randomize color
        for geoms that have no material.
        """
        self.whiten_materials()
        for name in self.geom_names:
            if self._check_geom_for_texture(name):
                # geom has valid texture that can be randomized
                self._randomize_texture(name)
                # randomize material if requested
                if self.randomize_material:
                    self._randomize_material(name)
            else:
                # randomize geom color
                self._randomize_geom_color(name)

        if self.randomize_skybox:
            self._randomize_texture("skybox")

    def _randomize_geom_color(self, name):
        """
        Helper function to randomize color of a specific geom

        Args:
            name (str): Name of the geom to randomize for
        """
        if self.randomize_local:
            random_color = self.random_state.uniform(0, 1, size=3)
            rgb = (1.0 - self.local_rgb_interpolation) * self._defaults[name][
                "rgb"
            ] + self.local_rgb_interpolation * random_color
        else:
            rgb = self.random_state.uniform(0, 1, size=3)
        self.set_geom_rgb(name, rgb)

    def _randomize_texture(self, name):
        """
        Helper function to randomize texture of a specific geom

        Args:
            name (str): Name of the geom to randomize for
        """
        keys = list(self._texture_variation_callbacks.keys())
        choice = keys[self.random_state.randint(len(keys))]
        self._texture_variation_callbacks[choice](name)

    def _randomize_material(self, name):
        """
        Helper function to randomize material of a specific geom

        Args:
            name (str): Name of the geom to randomize for
        """
        # Return immediately if this is the skybox
        if name == "skybox":
            return
        # Grab material id
        mat_id = self._name_to_mat_id(name)
        # Randomize reflectance, shininess, and specular
        material = self.random_state.uniform(0, 1, size=3)  # (reflectance, shininess, specular)
        self.set_material(name, material, perturb=self.randomize_local)

    def rand_checker(self, name):
        """
        Generates a random checker pattern for a specific geom

        Args:
            name (str): Name of the geom to randomize for
        """
        rgb1, rgb2 = self.get_rand_rgb(2)
        self.set_checker(name, rgb1, rgb2, perturb=self.randomize_local)

    def rand_gradient(self, name):
        """
        Generates a random gradient pattern for a specific geom

        Args:
            name (str): Name of the geom to randomize for
        """
        rgb1, rgb2 = self.get_rand_rgb(2)
        vertical = bool(self.random_state.uniform() > 0.5)
        self.set_gradient(name, rgb1, rgb2, vertical=vertical, perturb=self.randomize_local)

    def rand_rgb(self, name):
        """
        Generates a random RGB color for a specific geom

        Args:
            name (str): Name of the geom to randomize for
        """
        rgb = self.get_rand_rgb()
        self.set_rgb(name, rgb, perturb=self.randomize_local)

    def rand_noise(self, name):
        """
        Generates a random RGB noise pattern for a specific geom

        Args:
            name (str): Name of the geom to randomize for
        """
        fraction = 0.1 + self.random_state.uniform() * 0.8
        rgb1, rgb2 = self.get_rand_rgb(2)
        self.set_noise(name, rgb1, rgb2, fraction, perturb=self.randomize_local)

    def whiten_materials(self):
        """
        Extends modder.TextureModder to also whiten geom_rgba

        Helper method for setting all material colors to white, otherwise
        the texture modifications won't take full effect.
        """
        for name in self.geom_names:
            # whiten geom
            geom_id = self.model.geom_name2id(name)
            self.model.geom_rgba[geom_id, :] = 1.0

            if self._check_geom_for_texture(name):
                # whiten material
                mat_id = self.model.geom_matid[geom_id]
                self.model.mat_rgba[mat_id, :] = 1.0

    def get_geom_rgb(self, name):
        """
        Grabs rgb color of a specific geom

        Args:
            name (str): Name of the geom

        Returns:
            np.array: (r,g,b) geom colors
        """
        geom_id = self.model.geom_name2id(name)
        return self.model.geom_rgba[geom_id, :3]

    def set_geom_rgb(self, name, rgb):
        """
        Sets rgb color of a specific geom

        Args:
            name (str): Name of the geom
            rgb (np.array): (r,g,b) geom colors
        """
        geom_id = self.model.geom_name2id(name)
        self.model.geom_rgba[geom_id, :3] = rgb

    def get_rand_rgb(self, n=1):
        """
        Grabs a batch of random rgb tuple combos

        Args:
            n (int): How many sets of rgb tuples to randomly generate

        Returns:
            np.array or n-tuple: if n > 1, each tuple entry is a rgb tuple. else, single (r,g,b) array
        """

        def _rand_rgb():
            return np.array(self.random_state.uniform(size=3) * 255, dtype=np.uint8)

        if n == 1:
            return _rand_rgb()
        else:
            return tuple(_rand_rgb() for _ in range(n))

    def get_texture(self, name):
        """
        Grabs texture of a specific geom

        Args:
            name (str): Name of the geom

        Returns:
            Texture: texture associated with the geom
        """
        tex_id = self._name_to_tex_id(name)
        texture = self.textures[tex_id]
        return texture

    def set_texture(self, name, bitmap, perturb=False):
        """
        Sets the bitmap for the texture that corresponds
        to geom @name.

        If @perturb is True, then use the computed bitmap
        to perturb the default bitmap slightly, instead
        of replacing it.

        Args:
            name (str): Name of the geom
            bitmap (np.array): 3d-array representing rgb pixel-wise values
            perturb (bool): Whether to perturb the inputted bitmap or not
        """
        bitmap_to_set = self.get_texture(name).bitmap
        if perturb:
            bitmap = (1.0 - self.local_rgb_interpolation) * self._defaults[name][
                "texture"
            ] + self.local_rgb_interpolation * bitmap
        bitmap_to_set[:] = bitmap
        self.upload_texture(name)

    def get_material(self, name):
        """
        Grabs material of a specific geom

        Args:
            name (str): Name of the geom

        Returns:
            np.array: (reflectance, shininess, specular) material properties associated with the geom
        """
        mat_id = self._name_to_mat_id(name)
        # Material is in tuple form (reflectance, shininess, specular)
        material = np.array(
            (self.model.mat_reflectance[mat_id], self.model.mat_shininess[mat_id], self.model.mat_specular[mat_id])
        )
        return material

    def set_material(self, name, material, perturb=False):
        """
        Sets the material that corresponds to geom @name.

        If @perturb is True, then use the computed material
        to perturb the default material slightly, instead
        of replacing it.

        Args:
            name (str): Name of the geom
            material (np.array): (reflectance, shininess, specular) material properties associated with the geom
            perturb (bool): Whether to perturb the inputted material properties or not
        """
        mat_id = self._name_to_mat_id(name)
        if perturb:
            material = (1.0 - self.local_material_interpolation) * self._defaults[name][
                "material"
            ] + self.local_material_interpolation * material
        self.model.mat_reflectance[mat_id] = material[0]
        self.model.mat_shininess[mat_id] = material[1]
        self.model.mat_specular[mat_id] = material[2]

    def get_checker_matrices(self, name):
        """
        Grabs checker pattern matrix associated with @name.

        Args:
            name (str): Name of geom

        Returns:
            np.array: 3d-array representing rgb checker pattern
        """
        tex_id = self._name_to_tex_id(name)
        return self._texture_checker_mats[tex_id]

    def set_checker(self, name, rgb1, rgb2, perturb=False):
        """
        Use the two checker matrices to create a checker
        pattern from the two colors, and set it as
        the texture for geom @name.

        Args:
            name (str): Name of geom
            rgb1 (3-array): (r,g,b) value for one half of checker pattern
            rgb2 (3-array): (r,g,b) value for other half of checker pattern
            perturb (bool): Whether to perturb the resulting checker pattern or not
        """
        cbd1, cbd2 = self.get_checker_matrices(name)
        rgb1 = np.asarray(rgb1).reshape([1, 1, -1])
        rgb2 = np.asarray(rgb2).reshape([1, 1, -1])
        bitmap = rgb1 * cbd1 + rgb2 * cbd2

        self.set_texture(name, bitmap, perturb=perturb)

    def set_gradient(self, name, rgb1, rgb2, vertical=True, perturb=False):
        """
        Creates a linear gradient from rgb1 to rgb2.

        Args:
            name (str): Name of geom
            rgb1 (3-array): start color
            rgb2 (3- array): end color
            vertical (bool): if True, the gradient in the positive
                y-direction, if False it's in the positive x-direction.
            perturb (bool): Whether to perturb the resulting gradient pattern or not
        """
        # NOTE: MuJoCo's gradient uses a sigmoid. Here we simplify
        # and just use a linear gradient... We could change this
        # to just use a tanh-sigmoid if needed.
        bitmap = self.get_texture(name).bitmap
        h, w = bitmap.shape[:2]
        if vertical:
            p = np.tile(np.linspace(0, 1, h)[:, None], (1, w))
        else:
            p = np.tile(np.linspace(0, 1, w), (h, 1))

        new_bitmap = np.zeros_like(bitmap)
        for i in range(3):
            new_bitmap[..., i] = rgb2[i] * p + rgb1[i] * (1.0 - p)

        self.set_texture(name, new_bitmap, perturb=perturb)

    def set_rgb(self, name, rgb, perturb=False):
        """
        Just set the texture bitmap for geom @name
        to a constant rgb value.

        Args:
            name (str): Name of geom
            rgb (3-array): desired (r,g,b) color
            perturb (bool): Whether to perturb the resulting color pattern or not
        """
        bitmap = self.get_texture(name).bitmap
        new_bitmap = np.zeros_like(bitmap)
        new_bitmap[..., :] = np.asarray(rgb)

        self.set_texture(name, new_bitmap, perturb=perturb)

    def set_noise(self, name, rgb1, rgb2, fraction=0.9, perturb=False):
        """
        Sets the texture bitmap for geom @name to a noise pattern

        Args:
            name (str): name of geom
            rgb1 (3-array): background color
            rgb2 (3-array): color of random noise foreground color
            fraction (float): fraction of pixels with foreground color
            perturb (bool): Whether to perturb the resulting color pattern or not
        """
        bitmap = self.get_texture(name).bitmap
        h, w = bitmap.shape[:2]
        mask = self.random_state.uniform(size=(h, w)) < fraction

        new_bitmap = np.zeros_like(bitmap)
        new_bitmap[..., :] = np.asarray(rgb1)
        new_bitmap[mask, :] = np.asarray(rgb2)

        self.set_texture(name, new_bitmap, perturb=perturb)

    def upload_texture(self, name, device_id=0):
        """
        Uploads the texture to the GPU so it's available in the rendering.

        Args:
            name (str): name of geom
        """
        texture = self.get_texture(name)
        if self.sim._render_context_offscreen is None:
            render_context = MjRenderContextOffscreen(self.sim, device_id)
            render_context.upload_texture(texture.id)

    def _check_geom_for_texture(self, name):
        """
        Helper function to determined if the geom @name has
        an assigned material and that the material has
        an assigned texture.

        Args:
            name (str): name of geom

        Returns:
            bool: True if specific geom has both material and texture associated, else False
        """
        geom_id = self.model.geom_name2id(name)
        mat_id = self.model.geom_matid[geom_id]
        if mat_id < 0:
            return False
        tex_id = self.model.mat_texid[mat_id]
        if tex_id < 0:
            return False
        return True

    def _name_to_tex_id(self, name):
        """
        Helper function to get texture id from geom name.

        Args:
            name (str): name of geom

        Returns:
            int: id of texture associated with geom

        Raises:
            AssertionError: [No texture associated with geom]
        """

        # handle skybox separately
        if name == "skybox":
            skybox_tex_id = -1
            for tex_id in range(self.model.ntex):
                skybox_textype = 2
                if self.model.tex_type[tex_id] == skybox_textype:
                    skybox_tex_id = tex_id
            assert skybox_tex_id >= 0
            return skybox_tex_id

        assert self._check_geom_for_texture(name)
        geom_id = self.model.geom_name2id(name)
        mat_id = self.model.geom_matid[geom_id]
        tex_id = self.model.mat_texid[mat_id]
        return tex_id

    def _name_to_mat_id(self, name):
        """
        Helper function to get material id from geom name.

        Args:
            name (str): name of geom

        Returns:
            int: id of material associated with geom

        Raises:
            ValueError: [No material associated with skybox]
            AssertionError: [No material associated with geom]
        """

        # handle skybox separately
        if name == "skybox":
            raise ValueError("Error: skybox has no material!")

        assert self._check_geom_for_texture(name)
        geom_id = self.model.geom_name2id(name)
        mat_id = self.model.geom_matid[geom_id]
        return mat_id

    def _cache_checker_matrices(self):
        """
        Cache two matrices of the form [[1, 0, 1, ...],
                                        [0, 1, 0, ...],
                                        ...]
        and                            [[0, 1, 0, ...],
                                        [1, 0, 1, ...],
                                        ...]
        for each texture. To use for fast creation of checkerboard patterns
        """
        self._texture_checker_mats = []
        for tex_id in range(self.model.ntex):
            texture = self.textures[tex_id]
            h, w = texture.bitmap.shape[:2]
            self._texture_checker_mats.append(self._make_checker_matrices(h, w))

    def _make_checker_matrices(self, h, w):
        """
        Helper function to quickly generate binary matrices used to create checker patterns

        Args:
            h (int): Desired height of matrices
            w (int): Desired width of matrices

        Returns:
            2-tuple:

                - (np.array): 2d-array representing first half of checker matrix
                - (np.array): 2d-array representing second half of checker matrix
        """
        re = np.r_[((w + 1) // 2) * [0, 1]]
        ro = np.r_[((w + 1) // 2) * [1, 0]]
        cbd1 = np.expand_dims(np.row_stack(((h + 1) // 2) * [re, ro]), -1)[:h, :w]
        cbd2 = np.expand_dims(np.row_stack(((h + 1) // 2) * [ro, re]), -1)[:h, :w]
        return cbd1, cbd2


# From mjtTexture
MJT_TEXTURE_ENUM = ["2d", "cube", "skybox"]


class Texture:
    """
    Helper class for operating on the MuJoCo textures.

    Args:
        model (MjModel): Mujoco sim model
        tex_id (int): id of specific texture in mujoco sim
    """

    __slots__ = ["id", "type", "height", "width", "tex_adr", "tex_rgb"]

    def __init__(self, model, tex_id):
        self.id = tex_id
        self.type = MJT_TEXTURE_ENUM[model.tex_type[tex_id]]
        self.height = model.tex_height[tex_id]
        self.width = model.tex_width[tex_id]
        self.tex_adr = model.tex_adr[tex_id]
        self.tex_rgb = model.tex_rgb

    @property
    def bitmap(self):
        """
        Grabs color bitmap associated with this texture from the mujoco sim.

        Returns:
            np.array: 3d-array representing the rgb texture bitmap
        """
        size = self.height * self.width * 3
        data = self.tex_rgb[self.tex_adr : self.tex_adr + size]
        return data.reshape((self.height, self.width, 3))


class DynamicsModder(BaseModder):
    """
    Modder for various dynamics properties of the mujoco model, such as friction, damping, etc.
    This can be used to modify parameters stored in MjModel (ie friction, damping, etc.) as
    well as optimizer parameters stored in PyMjOption (i.e.: medium density, viscosity, etc.)
    To modify a parameter, use the parameter to be changed as a keyword argument to
    self.mod and the new value as the value for that argument. Supports arbitrary many
    modifications in a single step. Example use:
        sim = MjSim(...)
        modder = DynamicsModder(sim)
        modder.mod("element1_name", "attr1", new_value1)
        modder.mod("element2_name", "attr2", new_value2)
        ...
        modder.update()

    NOTE: It is necessary to perform modder.update() after performing all modifications to make sure
        the changes are propagated

    NOTE: A full list of supported randomizable parameters can be seen by calling modder.dynamics_parameters

    NOTE: When modifying parameters belonging to MjModel.opt (e.g.: density, viscosity), no name should
        be specified (set it as None in mod(...)). This is because opt does not have a name attribute
        associated with it

    Args:
        sim (MjSim): Mujoco sim instance

        random_state (RandomState): instance of np.random.RandomState

        randomize_density (bool): If True, randomizes global medium density

        randomize_viscosity (bool): If True, randomizes global medium viscosity

        density_perturbation_ratio (float): Relative (fraction) magnitude of default density randomization

        viscosity_perturbation_ratio:  Relative (fraction) magnitude of default viscosity randomization

        body_names (None or list of str): list of bodies to use for randomization. If not provided, all
            bodies in the model are randomized.

        randomize_position (bool): If True, randomizes body positions

        randomize_quaternion (bool): If True, randomizes body quaternions

        randomize_inertia (bool): If True, randomizes body inertias (only applicable for non-zero mass bodies)

        randomize_mass (bool): If True, randomizes body masses (only applicable for non-zero mass bodies)

        position_perturbation_size (float): Magnitude of body position randomization

        quaternion_perturbation_size (float): Magnitude of body quaternion randomization (angle in radians)

        inertia_perturbation_ratio (float): Relative (fraction) magnitude of body inertia randomization

        mass_perturbation_ratio (float): Relative (fraction) magnitude of body mass randomization

        geom_names (None or list of str): list of geoms to use for randomization. If not provided, all
            geoms in the model are randomized.

        randomize_friction (bool): If True, randomizes geom frictions

        randomize_solref (bool): If True, randomizes geom solrefs

        randomize_solimp (bool): If True, randomizes geom solimps

        friction_perturbation_ratio (float): Relative (fraction) magnitude of geom friction randomization

        solref_perturbation_ratio (float): Relative (fraction) magnitude of geom solref randomization

        solimp_perturbation_ratio (float): Relative (fraction) magnitude of geom solimp randomization

        joint_names (None or list of str): list of joints to use for randomization. If not provided, all
            joints in the model are randomized.

        randomize_stiffness (bool): If True, randomizes joint stiffnesses

        randomize_frictionloss (bool): If True, randomizes joint frictionlosses

        randomize_damping (bool): If True, randomizes joint dampings

        randomize_armature (bool): If True, randomizes joint armatures

        stiffness_perturbation_ratio (float): Relative (fraction) magnitude of joint stiffness randomization

        frictionloss_perturbation_size (float): Magnitude of joint frictionloss randomization

        damping_perturbation_size (float): Magnitude of joint damping randomization

        armature_perturbation_size (float): Magnitude of joint armature randomization
    """

    def __init__(
        self,
        sim,
        random_state=None,
        # Opt parameters
        randomize_density=True,
        randomize_viscosity=True,
        density_perturbation_ratio=0.1,
        viscosity_perturbation_ratio=0.1,
        # Body parameters
        body_names=None,
        randomize_position=True,
        randomize_quaternion=True,
        randomize_inertia=True,
        randomize_mass=True,
        position_perturbation_size=0.02,
        quaternion_perturbation_size=0.02,
        inertia_perturbation_ratio=0.02,
        mass_perturbation_ratio=0.02,
        # Geom parameters
        geom_names=None,
        randomize_friction=True,
        randomize_solref=True,
        randomize_solimp=True,
        friction_perturbation_ratio=0.1,
        solref_perturbation_ratio=0.1,
        solimp_perturbation_ratio=0.1,
        # Joint parameters
        joint_names=None,
        randomize_stiffness=True,
        randomize_frictionloss=True,
        randomize_damping=True,
        randomize_armature=True,
        stiffness_perturbation_ratio=0.1,
        frictionloss_perturbation_size=0.05,
        damping_perturbation_size=0.01,
        armature_perturbation_size=0.01,
    ):
        super().__init__(sim=sim, random_state=random_state)

        # Setup relevant values
        self.dummy_bodies = set()
        # Find all bodies that don't have any mass associated with them
        for body_name in self.sim.model.body_names:
            body_id = self.sim.model.body_name2id(body_name)
            if self.sim.model.body_mass[body_id] == 0:
                self.dummy_bodies.add(body_name)

        # Get all values to randomize
        self.body_names = list(self.sim.model.body_names) if body_names is None else body_names
        self.geom_names = list(self.sim.model.geom_names) if geom_names is None else geom_names
        self.joint_names = list(self.sim.model.joint_names) if joint_names is None else joint_names

        # Setup randomization settings
        # Each dynamics randomization group has its set of randomizable parameters, each of which has
        # its own settings ["randomize": whether its actively being randomized, "perturbation": the (potentially)
        # relative magnitude of the randomization to use, "type": either "ratio" or "size" (relative or absolute
        # perturbations), and "clip": (low, high) values to clip the final perturbed value by]
        self.opt_randomizations = {
            "density": {
                "randomize": randomize_density,
                "perturbation": density_perturbation_ratio,
                "type": "ratio",
                "clip": (0.0, np.inf),
            },
            "viscosity": {
                "randomize": randomize_viscosity,
                "perturbation": viscosity_perturbation_ratio,
                "type": "ratio",
                "clip": (0.0, np.inf),
            },
        }

        self.body_randomizations = {
            "position": {
                "randomize": randomize_position,
                "perturbation": position_perturbation_size,
                "type": "size",
                "clip": (-np.inf, np.inf),
            },
            "quaternion": {
                "randomize": randomize_quaternion,
                "perturbation": quaternion_perturbation_size,
                "type": "size",
                "clip": (-np.inf, np.inf),
            },
            "inertia": {
                "randomize": randomize_inertia,
                "perturbation": inertia_perturbation_ratio,
                "type": "ratio",
                "clip": (0.0, np.inf),
            },
            "mass": {
                "randomize": randomize_mass,
                "perturbation": mass_perturbation_ratio,
                "type": "ratio",
                "clip": (0.0, np.inf),
            },
        }

        self.geom_randomizations = {
            "friction": {
                "randomize": randomize_friction,
                "perturbation": friction_perturbation_ratio,
                "type": "ratio",
                "clip": (0.0, np.inf),
            },
            "solref": {
                "randomize": randomize_solref,
                "perturbation": solref_perturbation_ratio,
                "type": "ratio",
                "clip": (0.0, 1.0),
            },
            "solimp": {
                "randomize": randomize_solimp,
                "perturbation": solimp_perturbation_ratio,
                "type": "ratio",
                "clip": (0.0, np.inf),
            },
        }

        self.joint_randomizations = {
            "stiffness": {
                "randomize": randomize_stiffness,
                "perturbation": stiffness_perturbation_ratio,
                "type": "ratio",
                "clip": (0.0, np.inf),
            },
            "frictionloss": {
                "randomize": randomize_frictionloss,
                "perturbation": frictionloss_perturbation_size,
                "type": "size",
                "clip": (0.0, np.inf),
            },
            "damping": {
                "randomize": randomize_damping,
                "perturbation": damping_perturbation_size,
                "type": "size",
                "clip": (0.0, np.inf),
            },
            "armature": {
                "randomize": randomize_armature,
                "perturbation": armature_perturbation_size,
                "type": "size",
                "clip": (0.0, np.inf),
            },
        }

        # Store defaults so we don't loss track of the original (non-perturbed) values
        self.opt_defaults = None
        self.body_defaults = None
        self.geom_defaults = None
        self.joint_defaults = None
        self.save_defaults()

    def save_defaults(self):
        """
        Grabs the current values for all parameters in sim and stores them as default values
        """
        self.opt_defaults = {
            None: {  # no name associated with the opt parameters
                "density": self.sim.model.opt.density,
                "viscosity": self.sim.model.opt.viscosity,
            }
        }

        self.body_defaults = {}
        for body_name in self.sim.model.body_names:
            body_id = self.sim.model.body_name2id(body_name)
            self.body_defaults[body_name] = {
                "position": np.array(self.sim.model.body_pos[body_id]),
                "quaternion": np.array(self.sim.model.body_quat[body_id]),
                "inertia": np.array(self.sim.model.body_inertia[body_id]),
                "mass": self.sim.model.body_mass[body_id],
            }

        self.geom_defaults = {}
        for geom_name in self.sim.model.geom_names:
            geom_id = self.sim.model.geom_name2id(geom_name)
            self.geom_defaults[geom_name] = {
                "friction": np.array(self.sim.model.geom_friction[geom_id]),
                "solref": np.array(self.sim.model.geom_solref[geom_id]),
                "solimp": np.array(self.sim.model.geom_solimp[geom_id]),
            }

        self.joint_defaults = {}
        for joint_name in self.sim.model.joint_names:
            joint_id = self.sim.model.joint_name2id(joint_name)
            dof_idx = [i for i, v in enumerate(self.sim.model.dof_jntid) if v == joint_id]
            self.joint_defaults[joint_name] = {
                "stiffness": self.sim.model.jnt_stiffness[joint_id],
                "frictionloss": np.array(self.sim.model.dof_frictionloss[dof_idx]),
                "damping": np.array(self.sim.model.dof_damping[dof_idx]),
                "armature": np.array(self.sim.model.dof_armature[dof_idx]),
            }

    def restore_defaults(self):
        """
        Restores the default values curently saved in this modder
        """
        # Loop through all defaults and set the default value in sim
        for group_defaults in (self.opt_defaults, self.body_defaults, self.geom_defaults, self.joint_defaults):
            for name, defaults in group_defaults.items():
                for attr, default_val in defaults.items():
                    self.mod(name=name, attr=attr, val=default_val)

        # Make sure changes propagate in sim
        self.update()

    def randomize(self):
        """
        Randomizes all enabled dynamics parameters in the simulation
        """
        for group_defaults, group_randomizations, group_randomize_names in zip(
            (self.opt_defaults, self.body_defaults, self.geom_defaults, self.joint_defaults),
            (self.opt_randomizations, self.body_randomizations, self.geom_randomizations, self.joint_randomizations),
            ([None], self.body_names, self.geom_names, self.joint_names),
        ):
            for name in group_randomize_names:
                # Randomize all parameters associated with this element
                for attr, default_val in group_defaults[name].items():
                    val = copy.copy(default_val)
                    settings = group_randomizations[attr]
                    if settings["randomize"]:
                        # Randomize accordingly, and clip the final perturbed value
                        perturbation = np.random.rand() if type(val) in {int, float} else np.random.rand(*val.shape)
                        perturbation = settings["perturbation"] * (-1 + 2 * perturbation)
                        val = val + perturbation if settings["type"] == "size" else val * (1.0 + perturbation)
                        val = np.clip(val, *settings["clip"])
                    # Modify this value
                    self.mod(name=name, attr=attr, val=val)

        # Make sure changes propagate in sim
        self.update()

    def update_sim(self, sim):
        """
        In addition to super method, update internal default values to match the current values from
        (the presumably new) @sim.

        Args:
            sim (MjSim): MjSim object
        """
        super().update_sim(sim=sim)
        self.save_defaults()

    def update(self):
        """
        Propagates the changes made up to this point through the simulation
        """
        self.sim.forward()

    def mod(self, name, attr, val):
        """
        General method to modify dynamics parameter @attr to be new value @val, associated with element @name.

        Args:
            name (str): Name of element to modify parameter. This can be a body, geom, or joint name. If modifying
                an opt parameter, this should be set to None
            attr (str): Name of the dynamics parameter to modify. Valid options are self.dynamics_parameters
            val (int or float or n-array): New value(s) to set for the given dynamics parameter. The type of this
                argument should match the expected type for the given parameter.
        """
        # Make sure specified parameter is valid, and then modify it
        assert (
            attr in self.dynamics_parameters
        ), "Invalid dynamics parameter specified! Supported parameters are: {};" " requested: {}".format(
            self.dynamics_parameters, attr
        )
        # Modify the requested parameter (uses a clean way to programmatically call the appropriate method)
        getattr(self, f"mod_{attr}")(name, val)

    def mod_density(self, name=None, val=0.0):
        """
        Modifies the global medium density of the simulation.
        See http://www.mujoco.org/book/XMLreference.html#option for more details.

        Args:
            name (str): Name for this element. Should be left as None (opt has no name attribute)
            val (float): New density value.
        """
        # Make sure inputs are of correct form
        assert name is None, "No name should be specified if modding density!"

        # Modify this value
        self.sim.model.opt.density = val

    def mod_viscosity(self, name=None, val=0.0):
        """
        Modifies the global medium viscosity of the simulation.
        See http://www.mujoco.org/book/XMLreference.html#option for more details.

        Args:
            name (str): Name for this element. Should be left as None (opt has no name attribute)
            val (float): New viscosity value.
        """
        # Make sure inputs are of correct form
        assert name is None, "No name should be specified if modding density!"

        # Modify this value
        self.sim.model.opt.viscosity = val

    def mod_position(self, name, val=(0, 0, 0)):
        """
        Modifies the @name's relative body position within the simulation.
        See http://www.mujoco.org/book/XMLreference.html#body for more details.

        Args:
            name (str): Name for this element.
            val (3-array): New (x, y, z) relative position.
        """
        # Modify this value
        body_id = self.sim.model.body_name2id(name)
        self.sim.model.body_pos[body_id] = np.array(val)

    def mod_quaternion(self, name, val=(1, 0, 0, 0)):
        """
        Modifies the @name's relative body orientation (quaternion) within the simulation.
        See http://www.mujoco.org/book/XMLreference.html#body for more details.

        Note: This method automatically normalizes the inputted value.

        Args:
            name (str): Name for this element.
            val (4-array): New (w, x, y, z) relative quaternion.
        """
        # Normalize the inputted value
        val = np.array(val) / np.linalg.norm(val)
        # Modify this value
        body_id = self.sim.model.body_name2id(name)
        self.sim.model.body_quat[body_id] = val

    def mod_inertia(self, name, val):
        """
        Modifies the @name's relative body inertia within the simulation.
        See http://www.mujoco.org/book/XMLreference.html#body for more details.

        Args:
            name (str): Name for this element.
            val (3-array): New (ixx, iyy, izz) diagonal values in the inertia matrix.
        """
        # Modify this value if it's not a dummy body
        if name not in self.dummy_bodies:
            body_id = self.sim.model.body_name2id(name)
            self.sim.model.body_inertia[body_id] = np.array(val)

    def mod_mass(self, name, val):
        """
        Modifies the @name's mass within the simulation.
        See http://www.mujoco.org/book/XMLreference.html#body for more details.

        Args:
            name (str): Name for this element.
            val (float): New mass.
        """
        # Modify this value if it's not a dummy body
        if name not in self.dummy_bodies:
            body_id = self.sim.model.body_name2id(name)
            self.sim.model.body_mass[body_id] = val

    def mod_friction(self, name, val):
        """
        Modifies the @name's geom friction within the simulation.
        See http://www.mujoco.org/book/XMLreference.html#geom for more details.

        Args:
            name (str): Name for this element.
            val (3-array): New (sliding, torsional, rolling) friction values.
        """
        # Modify this value
        geom_id = self.sim.model.geom_name2id(name)
        self.sim.model.geom_friction[geom_id] = np.array(val)

    def mod_solref(self, name, val):
        """
        Modifies the @name's geom contact solver parameters within the simulation.
        See http://www.mujoco.org/book/modeling.html#CSolver for more details.

        Args:
            name (str): Name for this element.
            val (2-array): New (timeconst, dampratio) solref values.
        """
        # Modify this value
        geom_id = self.sim.model.geom_name2id(name)
        self.sim.model.geom_solref[geom_id] = np.array(val)

    def mod_solimp(self, name, val):
        """
        Modifies the @name's geom contact solver impedance parameters within the simulation.
        See http://www.mujoco.org/book/modeling.html#CSolver for more details.

        Args:
            name (str): Name for this element.
            val (5-array): New (dmin, dmax, width, midpoint, power) solimp values.
        """
        # Modify this value
        geom_id = self.sim.model.geom_name2id(name)
        self.sim.model.geom_solimp[geom_id] = np.array(val)

    def mod_stiffness(self, name, val):
        """
        Modifies the @name's joint stiffness within the simulation.
        See http://www.mujoco.org/book/XMLreference.html#joint for more details.

        NOTE: If the stiffness is already at 0, we IGNORE this value since a non-stiff joint (i.e.: free-turning)
            joint is fundamentally different than a stiffened joint)

        Args:
            name (str): Name for this element.
            val (float): New stiffness.
        """
        # Modify this value (only if there is stiffness to begin with)
        jnt_id = self.sim.model.joint_name2id(name)
        if self.sim.model.jnt_stiffness[jnt_id] != 0:
            self.sim.model.jnt_stiffness[jnt_id] = val

    def mod_frictionloss(self, name, val):
        """
        Modifies the @name's joint frictionloss within the simulation.
        See http://www.mujoco.org/book/XMLreference.html#joint for more details.

        NOTE: If the requested joint is a free joint, it will be ignored since it does not
            make physical sense to have friction loss associated with this joint (air drag / damping
            is already captured implicitly by the medium density / viscosity values)

        Args:
            name (str): Name for this element.
            val (float): New friction loss.
        """
        # Modify this value (only if it's not a free joint)
        jnt_id = self.sim.model.joint_name2id(name)
        if self.sim.model.jnt_type[jnt_id] != 0:
            dof_idx = [i for i, v in enumerate(self.sim.model.dof_jntid) if v == jnt_id]
            self.sim.model.dof_frictionloss[dof_idx] = val

    def mod_damping(self, name, val):
        """
        Modifies the @name's joint damping within the simulation.
        See http://www.mujoco.org/book/XMLreference.html#joint for more details.

        NOTE: If the requested joint is a free joint, it will be ignored since it does not
            make physical sense to have damping associated with this joint (air drag / damping
            is already captured implicitly by the medium density / viscosity values)

        Args:
            name (str): Name for this element.
            val (float): New damping.
        """
        # Modify this value (only if it's not a free joint)
        jnt_id = self.sim.model.joint_name2id(name)
        if self.sim.model.jnt_type[jnt_id] != 0:
            dof_idx = [i for i, v in enumerate(self.sim.model.dof_jntid) if v == jnt_id]
            self.sim.model.dof_damping[dof_idx] = val

    def mod_armature(self, name, val):
        """
        Modifies the @name's joint armature within the simulation.
        See http://www.mujoco.org/book/XMLreference.html#joint for more details.

        Args:
            name (str): Name for this element.
            val (float): New armature.
        """
        # Modify this value (only if it's not a free joint)
        jnt_id = self.sim.model.joint_name2id(name)
        if self.sim.model.jnt_type[jnt_id] != 0:
            dof_idx = [i for i, v in enumerate(self.sim.model.dof_jntid) if v == jnt_id]
            self.sim.model.dof_armature[dof_idx] = val

    @property
    def dynamics_parameters(self):
        """
        Returns:
            set: All dynamics parameters that can be randomized using this modder.
        """
        return {
            # Opt parameters
            "density",
            "viscosity",
            # Body parameters
            "position",
            "quaternion",
            "inertia",
            "mass",
            # Geom parameters
            "friction",
            "solref",
            "solimp",
            # Joint parameters
            "stiffness",
            "frictionloss",
            "damping",
            "armature",
        }

    @property
    def opt(self):
        """
        Returns:
             PyMjOption: MjModel sim options
        """
        return self.sim.model.opt



from ftplib import FTP, all_errors
import os
import tarfile
from tqdm import tqdm
import datetime
import re

def connect_to_ftp(ftp_host, ftp_user, ftp_password, ftp_port=21):
    """连接到 FTP 服务器并返回 FTP 对象。"""
    ftp = FTP()
    try:
        ftp.connect(ftp_host, ftp_port, timeout=10)
        ftp.login(user=ftp_user, passwd=ftp_password)
        print(f"成功连接到 {ftp_host}")
    except all_errors as e:
        print(f"连接失败：{e}")
        return None
    return ftp

def list_directory(ftp, directory):
    """列出远程 FTP 目录内容。"""
    try:
        files = ftp.nlst(directory)
        print(f"目录 {directory} 内容: {files}")
        return files
    except all_errors as e:
        print(f"列出目录错误：{e}")
        return []

def ensure_directory_exists(directory):
    """确保目录存在，如果不存在则创建。"""
    if not os.path.exists(directory):
        os.makedirs(directory, exist_ok=True)
        print(f"创建目录: {directory}")

def download_file(ftp, remote_file_path, local_file_path):
    """带有进度条的文件下载。"""
    try:
        ensure_directory_exists(os.path.dirname(local_file_path))

        # 获取文件大小
        total_size = ftp.size(remote_file_path)

        # 打开文件并开始下载
        with open(local_file_path, 'wb') as local_file:
            with tqdm(total=total_size, unit='B', unit_scale=True, desc=local_file_path) as pbar:
                def callback(data):
                    local_file.write(data)
                    pbar.update(len(data))  # 更新进度条

                # 下载文件
                ftp.retrbinary(f'RETR {remote_file_path}', callback)

        print(f"文件已下载到 {local_file_path}")

    except all_errors as e:
        print(f"下载过程中发生错误：{e}")

def extract_tar_xz(file_path, extract_to):
    """解压 tar.xz 文件到指定目录。"""
    try:
        ensure_directory_exists(extract_to)
        with tarfile.open(file_path, "r:xz") as tar:
            tar.extractall(path=extract_to)
        print(f"文件已解压到 {extract_to}")

    except Exception as e:
        print(f"解压错误：{e}")

def find_latest_version(files, year, month):
    """根据当前年月找到最接近的版本。"""
    pattern = re.compile(r'v(\d{4})_(\d{1,2})_x\.tar\.xz')
    versions = []

    # 提取文件名并匹配符合格式的版本
    for file in files:
        filename = os.path.basename(file)  # 去除路径，保留文件名
        match = pattern.match(filename)
        if match:
            file_year, file_month = int(match.group(1)), int(match.group(2))
            versions.append((file_year, file_month, filename))

    # 根据年月排序，找到最接近的版本
    versions.sort(key=lambda x: (x[0], x[1]), reverse=True)

    # 找到当前年月或最近的版本
    for v_year, v_month, filename in versions:
        if (v_year < year) or (v_year == year and v_month <= month):
            print(f"找到匹配的版本: {filename}")
            return filename

    print("未找到匹配的版本")
    return None

def delete_file(file_path):
    """删除指定文件。"""
    try:
        os.remove(file_path)
        print(f"压缩包 {file_path} 已删除")
    except Exception as e:
        print(f"删除文件时出错：{e}")

def main():
    # 获取当前年月
    now = datetime.datetime.now()
    year, month = now.year, now.month

    # FTP 服务器信息
    ftp_host = "47.116.64.88"
    ftp_user = "Download"  # FTP 用户名（对外客户使用）
    ftp_password = "openverse"  # FTP 密码
    ftp_port = 20080  # FTP 端口
    remote_directory = "/orca-studio-projects"  # 远程目录路径

    # 获取当前脚本的运行目录
    current_working_dir = os.path.dirname(os.path.abspath(__file__))
    
    # 定义解压目录为当前目录的同级目录：orca-studio-projects
    parent_dir = os.path.dirname(current_working_dir)  # 获取当前目录的上一级目录
    extract_folder = os.path.join(parent_dir, "orca-studio-projects")
    ensure_directory_exists(extract_folder)  # 确保解压目录存在

    print(f"本地下载目录: {extract_folder}")
    print(f"解压目录: {extract_folder}")

    # 连接到 FTP 服务器
    ftp = connect_to_ftp(ftp_host, ftp_user, ftp_password, ftp_port)
    if not ftp:
        return

    # 列出远程目录内容
    files = list_directory(ftp, remote_directory)

    # 找到最新的版本
    version = find_latest_version(files, year, month)
    if not version:
        print("没有找到符合条件的版本，程序退出。")
        ftp.quit()
        return

    # 下载文件
    local_path = os.path.join(extract_folder, version)  # 下载到解压目录
    remote_path = f"{remote_directory}/{version}"
    download_file(ftp, remote_path, local_path)

    # 解压文件
    extract_tar_xz(local_path, extract_folder)

    # 删除压缩包
    delete_file(local_path)

    # 关闭 FTP 连接
    try:
        ftp.quit()
        print("FTP 连接已关闭")
    except OSError as e:
        print(f"关闭 FTP 连接时出错：{e}")

if __name__ == "__main__":
    main()
